from __future__ import annotations

import argparse
import importlib
import importlib.util
import logging
import textwrap
from pathlib import Path
from typing import TYPE_CHECKING

from dissect.cstruct import types
from dissect.cstruct.cstruct import cstruct

if TYPE_CHECKING:
    from types import ModuleType

log = logging.getLogger(__name__)


def load_module(path: Path, base: Path) -> ModuleType | None:
    module = None
    try:
        relative_path = path.relative_to(base)
        module_tuple = (*relative_path.parent.parts, relative_path.stem)
        spec = importlib.util.spec_from_file_location(".".join(module_tuple), path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
    except Exception as e:
        log.warning("Unable to import %s", path)
        log.debug("Error while trying to import module %s", path, exc_info=e)

    return module


def generate_file_stub(path: Path, base: Path) -> str:
    tmp_module = load_module(path, base)
    if tmp_module is None or not hasattr(tmp_module, "cstruct"):
        return ""

    header = [
        "# Generated by cstruct-stubgen",
        "from typing import BinaryIO, Literal, TypeAlias, overload",
        "",
        "import dissect.cstruct as __cs__",
    ]
    body = []

    for name, obj in tmp_module.__dict__.items():
        if isinstance(obj, cstruct):
            stub = generate_cstruct_stub(obj, module_prefix="__cs__.", cls_name=f"_{name}")
            body.append(stub)

            if body[-1][-1] != "\n":
                body.append("")

            body.append(f"# Technically `{name}` is an instance of `_{name}`, but then we can't use it in type hints")
            body.append(f"{name}: TypeAlias = _{name}")
            body.append("")

    if not body:
        return ""

    return "\n".join([*header, "", "\n".join(body)])


def generate_cstruct_stub(cs: cstruct, module_prefix: str = "", cls_name: str = "cstruct") -> str:
    empty_cs = cstruct()

    cs_prefix = f"{cls_name}."
    header = [f"class {cls_name}({module_prefix}cstruct):"]
    body = []
    indent = " " * 4

    # Constants first
    for name, value in cs.consts.items():
        if name in empty_cs.consts:
            continue
        body.append(textwrap.indent(f"{name}: Literal[{value!r}] = ...", prefix=indent))

    defined_names = set()

    # Then typedefs
    for name, typedef in cs.typedefs.items():
        if name in empty_cs.typedefs:
            continue

        if typedef.__name__ in empty_cs.typedefs:
            stub = f"{name}: TypeAlias = {cs_prefix}{typedef.__name__}"
        elif typedef.__name__ in defined_names:
            # Create an alias to the type if we have already seen it before.
            stub = f"{name}: TypeAlias = {typedef.__name__}"
        elif issubclass(typedef, (types.Enum, types.Flag)):
            stub = generate_enum_stub(typedef, cs_prefix=cs_prefix, module_prefix=module_prefix)
        elif issubclass(typedef, types.Pointer):
            typehint = generate_typehint(typedef, prefix=cs_prefix, module_prefix=module_prefix)
            stub = f"{name}: TypeAlias = {typehint}"
        elif issubclass(typedef, types.Structure):
            stub = generate_structure_stub(typedef, cs_prefix=cs_prefix, module_prefix=module_prefix)
        elif issubclass(typedef, types.BaseType):
            stub = generate_generic_stub(typedef, cs_prefix=cs_prefix, module_prefix=module_prefix)
        elif isinstance(typedef, str):
            stub = f"{name}: TypeAlias = {typedef}"
        else:
            raise TypeError(f"Unknown typedef: {typedef}")

        defined_names.add(typedef.__name__)

        body.append(textwrap.indent(stub, prefix=indent))

    if not body:
        body.append(textwrap.indent("...", prefix=indent))

    return "\n".join(header + body)


def generate_typehint(
    type_: type[types.BaseType],
    prefix: str = "",
    module_prefix: str = "",
) -> str:
    if issubclass(type_, types.CharArray):
        return f"{module_prefix}CharArray"
    if issubclass(type_, types.WcharArray):
        return f"{module_prefix}WcharArray"
    if issubclass(type_, types.Pointer):
        return f"{module_prefix}Pointer[{generate_typehint(type_.type, prefix, module_prefix)}]"
    if issubclass(type_, types.Array):
        return f"{module_prefix}Array[{generate_typehint(type_.type, prefix, module_prefix)}]"
    return f"{prefix}{type_.__name__}"


def generate_generic_stub(
    type_: type[types.BaseType],
    name_prefix: str = "",
    cs_prefix: str = "",
    module_prefix: str = "",
) -> str:
    return f"class {name_prefix}{type_.__name__}({module_prefix}{type_.__base__.__name__}): ...\n"


def generate_enum_stub(
    enum: type[types.Enum | types.Flag],
    name_prefix: str = "",
    cs_prefix: str = "",
    module_prefix: str = "",
) -> str:
    result = [f"class {name_prefix}{enum.__name__}({module_prefix}{enum.__base__.__name__}):"]
    result.extend(f"    {key} = ..." for key in enum.__members__)
    result.append("")

    return "\n".join(result)


def generate_structure_stub(
    structure: type[types.Structure],
    name_prefix: str = "",
    cs_prefix: str = "",
    module_prefix: str = "",
) -> str:
    result = [f"class {name_prefix}{structure.__name__}({module_prefix}{structure.__base__.__name__}):"]

    indent = " " * 4

    args = ["self"]
    for field_name, field in structure.fields.items():
        type_name = field.type.__name__
        inlined = False

        # If it's a structure and not globally defined, add an inline stub for it
        nested_type = field.type
        while issubclass(nested_type, types.BaseArray):
            nested_type = nested_type.type

        if issubclass(nested_type, types.Structure) and type_name not in structure.cs.typedefs:
            inlined = True
            inline_stub = generate_structure_stub(nested_type, cs_prefix=cs_prefix, module_prefix=module_prefix)

            result.append(textwrap.indent(inline_stub, prefix=indent))

        type_hint = generate_typehint(field.type, "" if inlined else f"{cs_prefix}", module_prefix)
        result.append(f"    {field_name}: {type_hint}")

        args.append(f"{field_name}: {type_hint} | None = ...")

    result.append(textwrap.indent("@overload", prefix=indent))
    result.append(textwrap.indent(f"def __init__({', '.join(args)}): ...", prefix=indent))
    result.append(textwrap.indent("@overload", prefix=indent))
    result.append(
        textwrap.indent("def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...", prefix=indent)
    )
    result.append("")
    return "\n".join(result)


def setup_logger(verbosity: int) -> None:
    level = logging.INFO
    if verbosity >= 1:
        level = logging.DEBUG

    logging.basicConfig(level=level)


def main() -> None:
    parser = argparse.ArgumentParser(
        "cstruct-stubify",
        description="Create .pyi stub files for cstruct definitions",
        epilog="NOTE: This tool will only generate stubs for the cstruct definitions in a file, not any other Python code. Manual fixups may be required.",  # noqa: E501
    )
    parser.add_argument("path", type=Path, help="path to the file or directory to create stubs for")
    parser.add_argument("-v", "--verbose", action="count", default=0)
    args = parser.parse_args()

    setup_logger(args.verbose)

    path: Path = args.path
    for file in path.rglob("*.py") if path.is_dir() else [path]:
        if file.is_file() and file.suffix == ".py":
            stub = generate_file_stub(file, path)
            if not stub:
                continue

            stub_file = file.with_suffix(".pyi")
            log.info("Writing stub of file %s to %s", file, stub_file.name)
            stub_file.write_text(stub)


if __name__ == "__main__":
    main()



from __future__ import annotations

from typing import TYPE_CHECKING, Any, BinaryIO

from dissect.cstruct.types.base import BaseType

if TYPE_CHECKING:
    from typing_extensions import Self


class LEB128(int, BaseType):
    """Variable-length code compression to store an arbitrarily large integer in a small number of bytes.

    See https://en.wikipedia.org/wiki/LEB128 for more information and an explanation of the algorithm.
    """

    signed: bool

    @classmethod
    def _read(cls, stream: BinaryIO, context: dict[str, Any] | None = None) -> Self:
        result = 0
        shift = 0
        while True:
            b = stream.read(1)
            if b == b"":
                raise EOFError("EOF reached, while final LEB128 byte was not yet read")

            b = ord(b)
            result |= (b & 0x7F) << shift
            shift += 7
            if (b & 0x80) == 0:
                break

        if cls.signed and b & 0x40 != 0:
            result |= ~0 << shift

        return cls.__new__(cls, result)

    @classmethod
    def _read_0(cls, stream: BinaryIO, context: dict[str, Any] | None = None) -> list[Self]:
        result = []

        while True:
            if (value := cls._read(stream, context)) == 0:
                break

            result.append(value)

        return result

    @classmethod
    def _write(cls, stream: BinaryIO, data: int) -> int:
        # only write negative numbers when in signed mode
        if data < 0 and not cls.signed:
            raise ValueError("Attempt to encode a negative integer using unsigned LEB128 encoding")

        result = bytearray()
        while True:
            # low-order 7 bits of value
            byte = data & 0x7F
            data = data >> 7

            # function works similar for signed- and unsigned integers, except for the check when to stop
            # the encoding process.
            if ((cls.signed and (data == 0 and byte & 0x40 == 0)) or (data == -1 and byte & 0x40 != 0)) or (
                not cls.signed and data == 0
            ):
                result.append(byte)
                break

            # Set high-order bit of byte
            result.append(0x80 | byte)

        stream.write(result)
        return len(result)



from __future__ import annotations

import ctypes as _ctypes
import struct
import sys
import types
from pathlib import Path
from typing import TYPE_CHECKING, Any, BinaryIO, TypeVar, cast

from dissect.cstruct.exceptions import ResolveError
from dissect.cstruct.expression import Expression
from dissect.cstruct.parser import CStyleParser, TokenParser
from dissect.cstruct.types import (
    LEB128,
    Array,
    BaseArray,
    BaseType,
    Char,
    Enum,
    Field,
    Flag,
    Int,
    Packed,
    Pointer,
    Structure,
    Union,
    Void,
    Wchar,
)

if TYPE_CHECKING:
    from collections.abc import Iterable
    from typing import TypeAlias


T = TypeVar("T", bound=BaseType)


class cstruct:
    """Main class of cstruct. All types are registered in here.

    Args:
        endian: The endianness to use when parsing.
        pointer: The pointer type to use for pointers.
    """

    DEF_CSTYLE = 1
    DEF_LEGACY = 2

    def __init__(self, load: str = "", *, endian: str = "<", pointer: str | None = None):
        self.endian = endian

        self.consts = {}
        self.lookups = {}
        self.includes = []
        # fmt: off
        self.typedefs = {
            # Internal types
            "int8": self._make_packed_type("int8", "b", int),
            "uint8": self._make_packed_type("uint8", "B", int),
            "int16": self._make_packed_type("int16", "h", int),
            "uint16": self._make_packed_type("uint16", "H", int),
            "int32": self._make_packed_type("int32", "i", int),
            "uint32": self._make_packed_type("uint32", "I", int),
            "int64": self._make_packed_type("int64", "q", int),
            "uint64": self._make_packed_type("uint64", "Q", int),
            "float16": self._make_packed_type("float16", "e", float),
            "float": self._make_packed_type("float", "f", float),
            "double": self._make_packed_type("double", "d", float),
            "char": self._make_type("char", (Char,), 1),
            "wchar": self._make_type("wchar", (Wchar,), 2),

            "int24": self._make_int_type("int24", 3, True, alignment=4),
            "uint24": self._make_int_type("uint24", 3, False, alignment=4),
            "int48": self._make_int_type("int48", 6, True, alignment=8),
            "uint48": self._make_int_type("int48", 6, False, alignment=8),
            "int128": self._make_int_type("int128", 16, True, alignment=16),
            "uint128": self._make_int_type("uint128", 16, False, alignment=16),

            "uleb128": self._make_type("uleb128", (LEB128,), None, attrs={"signed": False}),
            "ileb128": self._make_type("ileb128", (LEB128,), None, attrs={"signed": True}),

            "void": self._make_type("void", (Void,), 0),

            # Common C types not covered by internal types
            "signed char": "int8",
            "unsigned char": "char",
            "short": "int16",
            "signed short": "int16",
            "unsigned short": "uint16",
            "int": "int32",
            "signed int": "int32",
            "unsigned int": "uint32",
            "long": "int32",
            "signed long": "int32",
            "unsigned long": "uint32",
            "long long": "int64",
            "signed long long": "int64",
            "unsigned long long": "uint64",

            # Windows types
            "BYTE": "uint8",
            "CHAR": "char",
            "SHORT": "int16",
            "WORD": "uint16",
            "DWORD": "uint32",
            "LONG": "int32",
            "LONG32": "int32",
            "LONG64": "int64",
            "LONGLONG": "int64",
            "QWORD": "uint64",
            "OWORD": "uint128",
            "WCHAR": "wchar",

            "UCHAR": "uint8",
            "USHORT": "uint16",
            "ULONG": "uint32",
            "ULONG64": "uint64",
            "ULONGLONG": "uint64",

            "INT": "int32",
            "INT8": "int8",
            "INT16": "int16",
            "INT32": "int32",
            "INT64": "int64",
            "INT128": "int128",

            "UINT": "uint32",
            "UINT8": "uint8",
            "UINT16": "uint16",
            "UINT32": "uint32",
            "UINT64": "uint64",
            "UINT128": "uint128",

            "__int8": "int8",
            "__int16": "int16",
            "__int32": "int32",
            "__int64": "int64",
            "__int128": "int128",

            "unsigned __int8": "uint8",
            "unsigned __int16": "uint16",
            "unsigned __int32": "uint32",
            "unsigned __int64": "uint64",
            "unsigned __int128": "uint128",

            "wchar_t": "wchar",

            # GNU C types
            "int8_t": "int8",
            "int16_t": "int16",
            "int32_t": "int32",
            "int64_t": "int64",
            "int128_t": "int128",

            "uint8_t": "uint8",
            "uint16_t": "uint16",
            "uint32_t": "uint32",
            "uint64_t": "uint64",
            "uint128_t": "uint128",

            # IDA types
            "_BYTE": "uint8",
            "_WORD": "uint16",
            "_DWORD": "uint32",
            "_QWORD": "uint64",
            "_OWORD": "uint128",

            # Other convenience types
            "u1": "uint8",
            "u2": "uint16",
            "u4": "uint32",
            "u8": "uint64",
            "u16": "uint128",
            "__u8": "uint8",
            "__u16": "uint16",
            "__u32": "uint32",
            "__u64": "uint64",
            "uchar": "uint8",
            "ushort": "uint16",
            "uint": "uint32",
            "ulong": "uint32",
        }
        # fmt: on

        pointer = pointer or ("uint64" if sys.maxsize > 2**32 else "uint32")
        self.pointer: type[BaseType] = self.resolve(pointer)
        self._anonymous_count = 0

        if load:
            self.load(load)

    def __getattr__(self, attr: str) -> Any:
        try:
            return self.consts[attr]
        except KeyError:
            pass

        try:
            return self.resolve(self.typedefs[attr])
        except KeyError:
            pass

        raise AttributeError(f"Invalid attribute: {attr}")

    def _next_anonymous(self) -> str:
        name = f"__anonymous_{self._anonymous_count}__"
        self._anonymous_count += 1
        return name

    def add_type(self, name: str, type_: type[BaseType] | str, replace: bool = False) -> None:
        """Add a type or type reference.

        Only use this method when creating type aliases or adding already bound types.

        Args:
            name: Name of the type to be added.
            type_: The type to be added. Can be a str reference to another type or a compatible type class.

        Raises:
            ValueError: If the type already exists.
        """
        if not replace and (name in self.typedefs and self.resolve(self.typedefs[name]) != self.resolve(type_)):
            raise ValueError(f"Duplicate type: {name}")

        self.typedefs[name] = type_

    addtype = add_type

    def add_custom_type(
        self, name: str, type_: type[BaseType], size: int | None = None, alignment: int | None = None, **kwargs
    ) -> None:
        """Add a custom type.

        Use this method to add custom types to this cstruct instance. This is largely a convenience method for
        the internal :func:`_make_type` method, which binds a class to this cstruct instance.

        Args:
            name: Name of the type to be added.
            type_: The type to be added.
            size: The size of the type.
            alignment: The alignment of the type.
            **kwargs: Additional attributes to add to the type.
        """
        self.add_type(name, self._make_type(name, (type_,), size, alignment=alignment, attrs=kwargs))

    def load(self, definition: str, deftype: int | None = None, **kwargs) -> cstruct:
        """Parse structures from the given definitions using the given definition type.

        Definitions can be parsed using different parsers. Currently, there's
        only one supported parser - DEF_CSTYLE. Parsers can add types and
        modify this cstruct instance. Arguments can be passed to parsers
        using kwargs.

        The CSTYLE parser was recently replaced with token based parser,
        instead of a strictly regex based one. The old parser is still available
        by using DEF_LEGACY.

        Args:
            definition: The definition to parse.
            deftype: The definition type to parse the definitions with.
            **kwargs: Keyword arguments for parsers.
        """
        deftype = deftype or cstruct.DEF_CSTYLE

        if deftype == cstruct.DEF_CSTYLE:
            TokenParser(self, **kwargs).parse(definition)
        elif deftype == cstruct.DEF_LEGACY:
            CStyleParser(self, **kwargs).parse(definition)

        return self

    def loadfile(self, path: str, deftype: int | None = None, **kwargs) -> None:
        """Load structure definitions from a file.

        The given path will be read and parsed using the :meth:`~cstruct.load` function.

        Args:
            path: The path to load definitions from.
            deftype: The definition type to parse the definitions with.
            **kwargs: Keyword arguments for parsers.
        """
        with Path(path).open() as fh:
            self.load(fh.read(), deftype, **kwargs)

    def read(self, name: str, stream: BinaryIO) -> Any:
        """Parse data using a given type.

        Args:
            name: Type name to read.
            stream: File-like object or byte string to parse.

        Returns:
            The parsed data.
        """
        return self.resolve(name).read(stream)

    def resolve(self, name: type[BaseType] | str) -> type[BaseType]:
        """Resolve a type name to get the actual type object.

        Types can be referenced using different names. When we want
        the actual type object, we need to resolve these references.

        Args:
            name: Type name to resolve.

        Returns:
            The resolved type object.

        Raises:
            ResolveError: If the type can't be resolved.
        """
        type_name = name
        if not isinstance(type_name, str):
            return type_name

        for _ in range(10):
            if type_name not in self.typedefs:
                raise ResolveError(f"Unknown type {name}")

            type_name = self.typedefs[type_name]

            if not isinstance(type_name, str):
                return type_name

        raise ResolveError(f"Recursion limit exceeded while resolving type {name}")

    def _make_type(
        self,
        name: str,
        bases: Iterable[object],
        size: int | None,
        *,
        alignment: int | None = None,
        attrs: dict[str, Any] | None = None,
    ) -> type[BaseType]:
        """Create a new type class bound to this cstruct instance.

        All types are created using this method. This method automatically binds the type to this cstruct instance.
        """
        attrs = attrs or {}
        attrs.update(
            {
                "cs": self,
                "size": size,
                "dynamic": size is None,
                "alignment": alignment or size,
            }
        )
        return types.new_class(name, bases, {}, lambda ns: ns.update(attrs))

    def _make_array(self, type_: T, num_entries: int | Expression | None) -> type[Array[T]]:
        null_terminated = False
        if num_entries is None:
            null_terminated = True
            size = None
        elif isinstance(num_entries, Expression) or type_.dynamic:
            size = None
        else:
            if type_.size is None:
                raise ValueError(f"Cannot create array of dynamic type: {type_.__name__}")
            size = num_entries * type_.size

        name = f"{type_.__name__}[]" if null_terminated else f"{type_.__name__}[{num_entries}]"

        bases = (type_.ArrayType,)

        attrs = {
            "type": type_,
            "num_entries": num_entries,
            "null_terminated": null_terminated,
        }

        return cast("type[Array]", self._make_type(name, bases, size, alignment=type_.alignment, attrs=attrs))

    def _make_int_type(self, name: str, size: int, signed: bool, *, alignment: int | None = None) -> type[Int]:
        return cast("type[Int]", self._make_type(name, (Int,), size, alignment=alignment, attrs={"signed": signed}))

    def _make_packed_type(self, name: str, packchar: str, base: type, *, alignment: int | None = None) -> type[Packed]:
        return cast(
            "type[Packed]",
            self._make_type(
                name,
                (base, Packed),
                struct.calcsize(packchar),
                alignment=alignment,
                attrs={"packchar": packchar},
            ),
        )

    def _make_enum(self, name: str, type_: type[BaseType], values: dict[str, int]) -> type[Enum]:
        return Enum(self, name, type_, values)

    def _make_flag(self, name: str, type_: type[BaseType], values: dict[str, int]) -> type[Flag]:
        return Flag(self, name, type_, values)

    def _make_pointer(self, target: type[BaseType]) -> type[Pointer]:
        return self._make_type(
            f"{target.__name__}*",
            (Pointer,),
            self.pointer.size,
            alignment=self.pointer.alignment,
            attrs={"type": target},
        )

    def _make_struct(
        self,
        name: str,
        fields: list[Field],
        *,
        align: bool = False,
        anonymous: bool = False,
        base: type[Structure] = Structure,
    ) -> type[Structure]:
        return self._make_type(
            name,
            (base,),
            None,
            attrs={
                "fields": fields,
                "__align__": align,
                "__anonymous__": anonymous,
            },
        )

    def _make_union(
        self, name: str, fields: list[Field], *, align: bool = False, anonymous: bool = False
    ) -> type[Structure]:
        return self._make_struct(name, fields, align=align, anonymous=anonymous, base=Union)

    if TYPE_CHECKING:
        # ruff: noqa: PYI042
        _int = int
        _float = float

        class int8(_int, Packed[_int]): ...

        class uint8(_int, Packed[_int]): ...

        class int16(_int, Packed[_int]): ...

        class uint16(_int, Packed[_int]): ...

        class int32(_int, Packed[_int]): ...

        class uint32(_int, Packed[_int]): ...

        class int64(_int, Packed[_int]): ...

        class uint64(_int, Packed[_int]): ...

        class float16(_float, Packed[_float]): ...

        class float(_float, Packed[_float]): ...

        class double(_float, Packed[_float]): ...

        class char(Char): ...

        class wchar(Wchar): ...

        class int24(Int): ...

        class uint24(Int): ...

        class int48(Int): ...

        class uint48(Int): ...

        class int128(Int): ...

        class uint128(Int): ...

        class uleb128(LEB128): ...

        class ileb128(LEB128): ...

        class void(Void): ...

        # signed char: TypeAlias = int8
        # signed char: TypeAlias = char
        short: TypeAlias = int16
        # signed short: TypeAlias = int16
        # unsigned short: TypeAlias = uint16
        int: TypeAlias = int32
        # signed int: TypeAlias = int32
        # unsigned int: TypeAlias = uint32
        long: TypeAlias = int32
        # signed long: TypeAlias = int32
        # unsigned long: TypeAlias = uint32
        # long long: TypeAlias = int64
        # signed long long: TypeAlias = int64
        # unsigned long long: TypeAlias = uint64

        BYTE: TypeAlias = uint8
        CHAR: TypeAlias = char
        SHORT: TypeAlias = int16
        WORD: TypeAlias = uint16
        DWORD: TypeAlias = uint32
        LONG: TypeAlias = int32
        LONG32: TypeAlias = int32
        LONG64: TypeAlias = int64
        LONGLONG: TypeAlias = int64
        QWORD: TypeAlias = uint64
        OWORD: TypeAlias = uint128
        WCHAR: TypeAlias = wchar

        UCHAR: TypeAlias = uint8
        USHORT: TypeAlias = uint16
        ULONG: TypeAlias = uint32
        ULONG64: TypeAlias = uint64
        ULONGLONG: TypeAlias = uint64

        INT: TypeAlias = int32
        INT8: TypeAlias = int8
        INT16: TypeAlias = int16
        INT32: TypeAlias = int32
        INT64: TypeAlias = int64
        INT128: TypeAlias = int128

        UINT: TypeAlias = uint32
        UINT8: TypeAlias = uint8
        UINT16: TypeAlias = uint16
        UINT32: TypeAlias = uint32
        UINT64: TypeAlias = uint64
        UINT128: TypeAlias = uint128

        __int8: TypeAlias = int8
        __int16: TypeAlias = int16
        __int32: TypeAlias = int32
        __int64: TypeAlias = int64
        __int128: TypeAlias = int128

        # unsigned __int8: TypeAlias = uint8
        # unsigned __int16: TypeAlias = uint16
        # unsigned __int32: TypeAlias = uint32
        # unsigned __int64: TypeAlias = uint64
        # unsigned __int128: TypeAlias = uint128

        wchar_t: TypeAlias = wchar

        int8_t: TypeAlias = int8
        int16_t: TypeAlias = int16
        int32_t: TypeAlias = int32
        int64_t: TypeAlias = int64
        int128_t: TypeAlias = int128

        uint8_t: TypeAlias = uint8
        uint16_t: TypeAlias = uint16
        uint32_t: TypeAlias = uint32
        uint64_t: TypeAlias = uint64
        uint128_t: TypeAlias = uint128

        _BYTE: TypeAlias = uint8
        _WORD: TypeAlias = uint16
        _DWORD: TypeAlias = uint32
        _QWORD: TypeAlias = uint64
        _OWORD: TypeAlias = uint128

        u1: TypeAlias = uint8
        u2: TypeAlias = uint16
        u4: TypeAlias = uint32
        u8: TypeAlias = uint64
        u16: TypeAlias = uint128
        __u8: TypeAlias = uint8
        __u16: TypeAlias = uint16
        __u32: TypeAlias = uint32
        __u64: TypeAlias = uint64
        uchar: TypeAlias = uint8
        ushort: TypeAlias = uint16
        uint: TypeAlias = uint32
        ulong: TypeAlias = uint32


def ctypes(structure: type[Structure]) -> type[_ctypes.Structure]:
    """Create ctypes structures from cstruct structures."""
    fields = []
    for field in structure.__fields__:
        t = ctypes_type(field.type)
        fields.append((field._name, t))

    return type(structure.__name__, (_ctypes.Structure,), {"_fields_": fields})


def ctypes_type(type_: type[BaseType]) -> Any:
    mapping = {
        "b": _ctypes.c_int8,
        "B": _ctypes.c_uint8,
        "h": _ctypes.c_int16,
        "H": _ctypes.c_uint16,
        "i": _ctypes.c_int32,
        "I": _ctypes.c_uint32,
        "q": _ctypes.c_int64,
        "Q": _ctypes.c_uint64,
        "f": _ctypes.c_float,
        "d": _ctypes.c_double,
    }

    if issubclass(type_, Packed) and type_.packchar in mapping:
        return mapping[type_.packchar]

    if issubclass(type_, Char):
        return _ctypes.c_char

    if issubclass(type_, Wchar):
        return _ctypes.c_wchar

    if issubclass(type_, BaseArray):
        subtype = ctypes_type(type_.type)
        return subtype * type_.num_entries

    if issubclass(type_, Pointer):
        subtype = ctypes_type(type_.type)
        return _ctypes.POINTER(subtype)

    if issubclass(type_, Structure):
        return ctypes(type_)

    raise NotImplementedError(f"Type not implemented: {type_.__class__.__name__}")



from __future__ import annotations

from typing import TYPE_CHECKING, Any, BinaryIO

from dissect.cstruct.types.base import BaseType
from dissect.cstruct.utils import ENDIANNESS_MAP

if TYPE_CHECKING:
    from typing_extensions import Self


class Int(int, BaseType):
    """Integer type that can span an arbitrary amount of bytes."""

    signed: bool

    @classmethod
    def _read(cls, stream: BinaryIO, context: dict[str, Any] | None = None) -> Self:
        data = stream.read(cls.size)

        if len(data) != cls.size:
            raise EOFError(f"Read {len(data)} bytes, but expected {cls.size}")

        return cls.from_bytes(data, ENDIANNESS_MAP[cls.cs.endian], signed=cls.signed)

    @classmethod
    def _read_0(cls, stream: BinaryIO, context: dict[str, Any] | None = None) -> Self:
        result = []

        while True:
            if (value := cls._read(stream, context)) == 0:
                break

            result.append(value)

        return result

    @classmethod
    def _write(cls, stream: BinaryIO, data: int) -> int:
        return stream.write(data.to_bytes(cls.size, ENDIANNESS_MAP[cls.cs.endian], signed=cls.signed))



from __future__ import annotations

from typing import TYPE_CHECKING, Any, BinaryIO

from dissect.cstruct.types.base import EOF, BaseArray, BaseType

if TYPE_CHECKING:
    from typing_extensions import Self


class CharArray(bytes, BaseArray):
    """Character array type for reading and writing byte strings."""

    @classmethod
    def __default__(cls) -> Self:
        return type.__call__(cls, b"\x00" * (0 if cls.dynamic or cls.null_terminated else cls.num_entries))

    @classmethod
    def _read(cls, stream: BinaryIO, context: dict[str, Any] | None = None) -> Self:
        return type.__call__(cls, super()._read(stream, context))

    @classmethod
    def _write(cls, stream: BinaryIO, data: bytes) -> int:
        if isinstance(data, list) and data and isinstance(data[0], int):
            data = bytes(data)

        elif isinstance(data, str):
            data = data.encode("latin-1")

        if cls.null_terminated:
            return stream.write(data + b"\x00")
        return stream.write(data)


class Char(bytes, BaseType):
    """Character type for reading and writing bytes."""

    ArrayType = CharArray

    @classmethod
    def __default__(cls) -> Self:
        return type.__call__(cls, b"\x00")

    @classmethod
    def _read(cls, stream: BinaryIO, context: dict[str, Any] | None = None) -> Self:
        return cls._read_array(stream, 1, context)

    @classmethod
    def _read_array(cls, stream: BinaryIO, count: int, context: dict[str, Any] | None = None) -> Self:
        if count == 0:
            return type.__call__(cls, b"")

        data = stream.read(-1 if count == EOF else count)
        if count != EOF and len(data) != count:
            raise EOFError(f"Read {len(data)} bytes, but expected {count}")

        return type.__call__(cls, data)

    @classmethod
    def _read_0(cls, stream: BinaryIO, context: dict[str, Any] | None = None) -> Self:
        buf = []
        while True:
            byte = stream.read(1)
            if byte == b"":
                raise EOFError("Read 0 bytes, but expected 1")

            if byte == b"\x00":
                break

            buf.append(byte)

        return type.__call__(cls, b"".join(buf))

    @classmethod
    def _write(cls, stream: BinaryIO, data: bytes | int | str) -> int:
        if isinstance(data, int):
            data = chr(data)

        if isinstance(data, str):
            data = data.encode("latin-1")

        return stream.write(data)



from __future__ import annotations

import string
from typing import TYPE_CHECKING, ClassVar

from dissect.cstruct.exceptions import ExpressionParserError, ExpressionTokenizerError

if TYPE_CHECKING:
    from collections.abc import Callable

    from dissect.cstruct import cstruct


HEXBIN_SUFFIX = {"x", "X", "b", "B"}


class ExpressionTokenizer:
    def __init__(self, expression: str):
        self.expression = expression
        self.pos = 0
        self.tokens = []

    def equal(self, token: str, expected: str | set[str]) -> bool:
        if isinstance(expected, set):
            return token in expected
        return token == expected

    def alnum(self, token: str) -> bool:
        return token.isalnum()

    def alpha(self, token: str) -> bool:
        return token.isalpha()

    def digit(self, token: str) -> bool:
        return token.isdigit()

    def hexdigit(self, token: str) -> bool:
        return token in string.hexdigits

    def operator(self, token: str) -> bool:
        return token in {"*", "/", "+", "-", "%", "&", "^", "|", "(", ")", "~"}

    def match(
        self,
        func: Callable[[str], bool] | None = None,
        expected: str | None = None,
        consume: bool = True,
        append: bool = True,
    ) -> bool:
        if self.eol():
            return False

        token = self.get_token()

        if expected and self.equal(token, expected):
            if append:
                self.tokens.append(token)
            if consume:
                self.consume()
            return True

        if func and func(token):
            if append:
                self.tokens.append(token)
            if consume:
                self.consume()
            return True

        return False

    def consume(self) -> None:
        self.pos += 1

    def eol(self) -> bool:
        return self.pos >= len(self.expression)

    def get_token(self) -> str:
        if self.eol():
            raise ExpressionTokenizerError(f"Out of bounds index: {self.pos}, length: {len(self.expression)}")
        return self.expression[self.pos]

    def tokenize(self) -> list[str]:
        token = ""

        # Loop over expression runs in linear time
        while not self.eol():
            # If token is a single character operand add it to tokens
            if self.match(self.operator):
                continue

            # If token is a single digit, keep looping over expression and build the number
            if self.match(self.digit, consume=False, append=False):
                token += self.get_token()
                self.consume()

                # Support for binary and hexadecimal notation
                if self.match(expected=HEXBIN_SUFFIX, consume=False, append=False):
                    token += self.get_token()
                    self.consume()

                while self.match(self.hexdigit, consume=False, append=False):
                    token += self.get_token()
                    self.consume()
                    if self.eol():
                        break

                # Checks for suffixes in numbers
                if self.match(expected={"u", "U"}, consume=False, append=False):
                    self.consume()
                    self.match(expected={"l", "L"}, append=False)
                    self.match(expected={"l", "L"}, append=False)

                elif self.match(expected={"l", "L"}, append=False):
                    self.match(expected={"l", "L"}, append=False)
                    self.match(expected={"u", "U"}, append=False)
                else:
                    pass

                # Number cannot end on x or b in the case of binary or hexadecimal notation
                if len(token) == 2 and token[-1] in HEXBIN_SUFFIX:
                    raise ExpressionTokenizerError("Invalid binary or hex notation")

                if len(token) > 1 and token[0] == "0" and token[1] not in HEXBIN_SUFFIX:
                    token = token[:1] + "o" + token[1:]
                self.tokens.append(token)
                token = ""

            # If token is alpha or underscore we need to build the identifier
            elif self.match(self.alpha, consume=False, append=False) or self.match(
                expected="_", consume=False, append=False
            ):
                while self.match(self.alnum, consume=False, append=False) or self.match(
                    expected="_", consume=False, append=False
                ):
                    token += self.get_token()
                    self.consume()
                    if self.eol():
                        break
                self.tokens.append(token)
                token = ""
            # If token is length 2 operand make sure next character is part of length 2 operand append to tokens
            elif self.match(expected=">", append=False) and self.match(expected=">", append=False):
                self.tokens.append(">>")
            elif self.match(expected="<", append=False) and self.match(expected="<", append=False):
                self.tokens.append("<<")
            elif self.match(expected={" ", "\n", "\t"}, append=False):
                continue
            else:
                raise ExpressionTokenizerError(
                    f"Tokenizer does not recognize following token '{self.expression[self.pos]}'"
                )
        return self.tokens


class Expression:
    """Expression parser for calculations in definitions."""

    binary_operators: ClassVar[dict[str, Callable[[int, int], int]]] = {
        "|": lambda a, b: a | b,
        "^": lambda a, b: a ^ b,
        "&": lambda a, b: a & b,
        "<<": lambda a, b: a << b,
        ">>": lambda a, b: a >> b,
        "+": lambda a, b: a + b,
        "-": lambda a, b: a - b,
        "*": lambda a, b: a * b,
        "/": lambda a, b: a // b,
        "%": lambda a, b: a % b,
    }

    unary_operators: ClassVar[dict[str, Callable[[int], int]]] = {
        "u": lambda a: -a,
        "~": lambda a: ~a,
    }

    precedence_levels: ClassVar[dict[str, int]] = {
        "|": 0,
        "^": 1,
        "&": 2,
        "<<": 3,
        ">>": 3,
        "+": 4,
        "-": 4,
        "*": 5,
        "/": 5,
        "%": 5,
        "u": 6,
        "~": 6,
        "sizeof": 6,
    }

    def __init__(self, expression: str):
        self.expression = expression
        self.tokens = ExpressionTokenizer(expression).tokenize()
        self.stack = []
        self.queue = []

    def __repr__(self) -> str:
        return self.expression

    def precedence(self, o1: str, o2: str) -> bool:
        return self.precedence_levels[o1] >= self.precedence_levels[o2]

    def evaluate_exp(self) -> None:
        operator = self.stack.pop(-1)
        res = 0

        if len(self.queue) < 1:
            raise ExpressionParserError("Invalid expression: not enough operands")

        right = self.queue.pop(-1)
        if operator in self.unary_operators:
            res = self.unary_operators[operator](right)
        else:
            if len(self.queue) < 1:
                raise ExpressionParserError("Invalid expression: not enough operands")

            left = self.queue.pop(-1)
            res = self.binary_operators[operator](left, right)

        self.queue.append(res)

    def is_number(self, token: str) -> bool:
        return token.isnumeric() or (len(token) > 2 and token[0] == "0" and token[1] in ("x", "X", "b", "B", "o", "O"))

    def evaluate(self, cs: cstruct, context: dict[str, int] | None = None) -> int:
        """Evaluates an expression using a Shunting-Yard implementation."""

        self.stack = []
        self.queue = []
        operators = set(self.binary_operators.keys()) | set(self.unary_operators.keys())

        context = context or {}
        tmp_expression = self.tokens

        # Unary minus tokens; we change the semantic of '-' depending on the previous token
        for i in range(len(self.tokens)):
            if self.tokens[i] == "-":
                if i == 0:
                    self.tokens[i] = "u"
                    continue
                if self.tokens[i - 1] in operators or self.tokens[i - 1] == "u" or self.tokens[i - 1] == "(":
                    self.tokens[i] = "u"
                    continue

        i = 0
        while i < len(tmp_expression):
            current_token = tmp_expression[i]
            if self.is_number(current_token):
                self.queue.append(int(current_token, 0))
            elif current_token in context:
                self.queue.append(int(context[current_token]))
            elif current_token in cs.consts:
                self.queue.append(int(cs.consts[current_token]))
            elif current_token in self.unary_operators:
                self.stack.append(current_token)
            elif current_token == "sizeof":
                if len(tmp_expression) < i + 3 or (tmp_expression[i + 1] != "(" or tmp_expression[i + 3] != ")"):
                    raise ExpressionParserError("Invalid sizeof operation")
                self.queue.append(len(cs.resolve(tmp_expression[i + 2])))
                i += 3
            elif current_token in operators:
                while (
                    len(self.stack) != 0 and self.stack[-1] != "(" and (self.precedence(self.stack[-1], current_token))
                ):
                    self.evaluate_exp()
                self.stack.append(current_token)
            elif current_token == "(":
                if i > 0:
                    previous_token = tmp_expression[i - 1]
                    if self.is_number(previous_token):
                        raise ExpressionParserError(
                            f"Parser expected sizeof or an arethmethic operator instead got: '{previous_token}'"
                        )

                self.stack.append(current_token)
            elif current_token == ")":
                if i > 0:
                    previous_token = tmp_expression[i - 1]
                    if previous_token == "(":
                        raise ExpressionParserError(
                            f"Parser expected an expression, instead received empty parenthesis. Index: {i}"
                        )

                if len(self.stack) == 0:
                    raise ExpressionParserError("Invalid expression")

                while self.stack[-1] != "(":
                    self.evaluate_exp()

                self.stack.pop(-1)
            else:
                raise ExpressionParserError(f"Unmatched token: '{current_token}'")
            i += 1

        while len(self.stack) != 0:
            if self.stack[-1] == "(":
                raise ExpressionParserError("Invalid expression")

            self.evaluate_exp()

        if len(self.queue) != 1:
            raise ExpressionParserError("Invalid expression")

        return self.queue[0]



from __future__ import annotations

import sys
from enum import Enum as _Enum
from enum import EnumMeta, IntEnum, IntFlag
from typing import TYPE_CHECKING, Any, BinaryIO, TypeVar, overload

from dissect.cstruct.types.base import Array, BaseType, MetaType

if TYPE_CHECKING:
    from typing_extensions import Self

    from dissect.cstruct.cstruct import cstruct


PY_311 = sys.version_info >= (3, 11, 0)
PY_312 = sys.version_info >= (3, 12, 0)

_S = TypeVar("_S")


class EnumMetaType(EnumMeta, MetaType):
    type: type[BaseType]

    @overload
    def __call__(cls, value: cstruct, name: str, type_: type[BaseType], *args, **kwargs) -> type[Enum]: ...

    @overload
    def __call__(cls: type[_S], value: int | BinaryIO | bytes) -> _S: ...

    def __call__(
        cls,
        value: cstruct | int | BinaryIO | bytes | None = None,
        name: str | None = None,
        type_: type[BaseType] | None = None,
        *args,
        **kwargs,
    ) -> Enum | type[Enum]:
        if name is None:
            if value is None:
                value = cls.type.__default__()

            if not isinstance(value, int):
                # value is a parsable value
                value = cls.type(value)

            return super().__call__(value)

        # We are constructing a new Enum class
        # cs is the cstruct instance, but we can't isinstance check it due to circular imports
        cs = value
        if not issubclass(type_, int):
            raise TypeError("Enum can only be created from int type")

        enum_cls = super().__call__(name, *args, **kwargs)
        enum_cls.cs = cs
        enum_cls.type = type_
        enum_cls.size = type_.size
        enum_cls.dynamic = type_.dynamic
        enum_cls.alignment = type_.alignment

        _fix_alias_members(enum_cls)

        return enum_cls

    @overload
    def __getitem__(cls: type[_S], name: str) -> _S: ...

    @overload
    def __getitem__(cls: type[_S], name: int) -> Array: ...

    def __getitem__(cls: type[_S], name: str | int) -> _S | Array:
        if isinstance(name, str):
            return super().__getitem__(name)
        return MetaType.__getitem__(cls, name)

    __len__ = MetaType.__len__

    def __contains__(cls, value: Any) -> bool:
        # We used to let stdlib enum handle `__contains__``` but this commit is incompatible with our API:
        # https://github.com/python/cpython/commit/8a9aee71268c77867d3cc96d43cbbdcbe8c0e1e8
        if isinstance(value, cls):
            return True
        return value in cls._value2member_map_

    def _read(cls, stream: BinaryIO, context: dict[str, Any] | None = None) -> Self:
        return cls(cls.type._read(stream, context))

    def _read_array(cls, stream: BinaryIO, count: int, context: dict[str, Any] | None = None) -> list[Self]:
        return list(map(cls, cls.type._read_array(stream, count, context)))

    def _read_0(cls, stream: BinaryIO, context: dict[str, Any] | None = None) -> list[Self]:
        return list(map(cls, cls.type._read_0(stream, context)))

    def _write(cls, stream: BinaryIO, data: Enum) -> int:
        return cls.type._write(stream, data.value)

    def _write_array(cls, stream: BinaryIO, array: list[BaseType | int]) -> int:
        data = [entry.value if isinstance(entry, _Enum) else entry for entry in array]
        return cls.type._write_array(stream, data)

    def _write_0(cls, stream: BinaryIO, array: list[BaseType | int]) -> int:
        data = [entry.value if isinstance(entry, _Enum) else entry for entry in array]
        return cls._write_array(stream, [*data, cls.type.__default__()])


def _fix_alias_members(cls: type[Enum]) -> None:
    # Emulate aenum NoAlias behaviour
    # https://github.com/ethanfurman/aenum/blob/master/aenum/doc/aenum.rst
    if len(cls._member_names_) == len(cls._member_map_):
        return

    for name, member in cls._member_map_.items():
        if name != member.name:
            new_member = int.__new__(cls, member.value)
            new_member._name_ = name
            new_member._value_ = member.value

            type.__setattr__(cls, name, new_member)
            cls._member_names_.append(name)
            cls._member_map_[name] = new_member
            cls._value2member_map_[member.value] = new_member


class Enum(BaseType, IntEnum, metaclass=EnumMetaType):
    """Enum type supercharged with cstruct functionality.

    Enums are (mostly) compatible with the Python 3 standard library ``IntEnum`` with some notable differences:
        - Duplicate members are their own unique member instead of being an alias
        - Non-existing values are allowed and handled similarly to ``IntFlag``: ``<Enum: 0>``
        - Enum members are only considered equal if the enum class is the same

    Enums can be made using any integer type.

    Example:
        When using the default C-style parser, the following syntax is supported::

            enum <name> [: <type>] {
                <values>
            };

        For example, an enum that has A=1, B=5 and C=6 could be written like so::

            enum Test : uint16 {
                A, B=5, C
            };
    """

    if PY_311:

        def __repr__(self) -> str:
            # Use the IntFlag repr as a base since it handles unknown values the way we want it
            # I.e. <Color: 255> instead of <Color.None: 255>
            result = IntFlag.__repr__(self)
            if not self.__class__.__name__:
                # Deal with anonymous enums by stripping off the first bit
                # I.e. <.RED: 1> -> <RED: 1>
                result = f"<{result[2:]}"
            return result

        def __str__(self) -> str:
            # We differentiate with standard Python enums in that we use a more descriptive str representation
            # Standard Python enums just use the integer value as str, we use EnumName.ValueName
            # In case of anonymous enums, we just use the ValueName
            # In case of unknown members, we use the integer value (in combination with the EnumName if there is one)
            base = f"{self.__class__.__name__}." if self.__class__.__name__ else ""
            value = self.name if self.name is not None else str(self.value)
            return f"{base}{value}"

    else:

        def __repr__(self) -> str:
            name = self.__class__.__name__
            if self._name_ is not None:
                if name:
                    name += "."
                name += self._name_
            return f"<{name}: {self._value_!r}>"

        def __str__(self) -> str:
            base = f"{self.__class__.__name__}." if self.__class__.__name__ else ""
            value = self._name_ if self._name_ is not None else str(self._value_)
            return f"{base}{value}"

    def __eq__(self, other: int | Enum) -> bool:
        if isinstance(other, Enum) and other.__class__ is not self.__class__:
            return False

        # Python <= 3.10 compatibility
        if isinstance(other, Enum):
            other = other.value

        return self.value == other

    def __ne__(self, value: int | Enum) -> bool:
        return not self.__eq__(value)

    def __hash__(self) -> int:
        return hash((self.__class__, self.name, self.value))

    @classmethod
    def _missing_(cls, value: int) -> Self:
        # Emulate FlagBoundary.KEEP for enum (allow values other than the defined members)
        new_member = int.__new__(cls, value)
        new_member._name_ = None
        new_member._value_ = value
        return new_member



from __future__ import annotations

from typing import TYPE_CHECKING, Any, BinaryIO

from dissect.cstruct.types.base import BaseArray, BaseType

if TYPE_CHECKING:
    from typing_extensions import Self


class VoidArray(list, BaseArray):
    """Array type representing void elements, primarily used for no-op reading and writing operations."""

    @classmethod
    def __default__(cls) -> Self:
        return cls()

    @classmethod
    def _read(cls, stream: BinaryIO, context: dict[str, Any] | None = None) -> Self:
        return cls()

    @classmethod
    def _write(cls, stream: BinaryIO, data: bytes) -> int:
        return 0


class Void(BaseType):
    """Void type."""

    ArrayType = VoidArray

    def __bool__(self) -> bool:
        return False

    def __eq__(self, value: object) -> bool:
        return isinstance(value, Void)

    @classmethod
    def _read(cls, stream: BinaryIO, context: dict[str, Any] | None = None) -> Self:
        return cls.__new__(cls)

    @classmethod
    def _write(cls, stream: BinaryIO, data: Void) -> int:
        return 0



from __future__ import annotations

import functools
from io import BytesIO
from typing import TYPE_CHECKING, Any, BinaryIO, ClassVar, TypeVar

from dissect.cstruct.exceptions import ArraySizeError
from dissect.cstruct.expression import Expression

if TYPE_CHECKING:
    from collections.abc import Callable

    from typing_extensions import Self

    from dissect.cstruct.cstruct import cstruct


EOF = -0xE0F  # Negative counts are illegal anyway, so abuse that for our EOF sentinel


class MetaType(type):
    """Base metaclass for cstruct type classes."""

    cs: cstruct
    """The cstruct instance this type class belongs to."""
    size: int | None
    """The size of the type in bytes. Can be ``None`` for dynamic sized types."""
    dynamic: bool
    """Whether or not the type is dynamically sized."""
    alignment: int | None
    """The alignment of the type in bytes. A value of ``None`` will be treated as 1-byte aligned."""

    # This must be the actual type, but since Array is a subclass of BaseType, we correct this at the bottom of the file
    ArrayType: type[BaseArray] = "Array"
    """The array type for this type class."""

    def __call__(cls, *args, **kwargs) -> Self:  # type: ignore
        """Adds support for ``TypeClass(bytes | file-like object)`` parsing syntax."""
        # TODO: add support for Type(cs) API to create new bounded type classes, similar to the old API?
        if len(args) == 1 and not isinstance(args[0], cls):
            stream = args[0]

            if _is_readable_type(stream):
                return cls._read(stream)

            if issubclass(cls, bytes) and isinstance(stream, bytes) and len(stream) == cls.size:
                # Shortcut for char/bytes type
                return type.__call__(cls, *args, **kwargs)

            if _is_buffer_type(stream):
                return cls.reads(stream)

        return type.__call__(cls, *args, **kwargs)

    def __getitem__(cls, num_entries: int | Expression | None) -> type[BaseArray]:
        """Create a new array with the given number of entries."""
        return cls.cs._make_array(cls, num_entries)

    def __bool__(cls) -> bool:
        """Type class is always truthy."""
        return True

    def __len__(cls) -> int:
        """Return the byte size of the type."""
        if cls.size is None:
            raise TypeError("Dynamic size")

        return cls.size

    def __default__(cls) -> Self:  # type: ignore
        """Return the default value of this type."""
        return cls()

    def reads(cls, data: bytes | memoryview | bytearray) -> Self:  # type: ignore
        """Parse the given data from a bytes-like object.

        Args:
            data: Bytes-like object to parse.

        Returns:
            The parsed value of this type.
        """
        return cls._read(BytesIO(data))

    def read(cls, obj: BinaryIO | bytes | memoryview | bytearray) -> Self:  # type: ignore
        """Parse the given data.

        Args:
            obj: Data to parse. Can be a bytes-like object or a file-like object.

        Returns:
            The parsed value of this type.
        """
        if _is_buffer_type(obj):
            return cls.reads(obj)

        if not _is_readable_type(obj):
            raise TypeError("Invalid object type")

        return cls._read(obj)

    def write(cls, stream: BinaryIO, value: Any) -> int:
        """Write a value to a writable file-like object.

        Args:
            stream: File-like objects that supports writing.
            value: Value to write.

        Returns:
            The amount of bytes written.
        """
        return cls._write(stream, value)

    def dumps(cls, value: Any) -> bytes:
        """Dump a value to a byte string.

        Args:
            value: Value to dump.

        Returns:
            The raw bytes of this type.
        """
        out = BytesIO()
        cls._write(out, value)
        return out.getvalue()

    def _read(cls, stream: BinaryIO, context: dict[str, Any] | None = None) -> Self:  # type: ignore
        """Internal function for reading value.

        Must be implemented per type.

        Args:
            stream: The stream to read from.
            context: Optional reading context.
        """
        raise NotImplementedError

    def _read_array(cls, stream: BinaryIO, count: int, context: dict[str, Any] | None = None) -> list[Self]:  # type: ignore
        """Internal function for reading array values.

        Allows type implementations to do optimized reading for their type.

        Args:
            stream: The stream to read from.
            count: The amount of values to read.
            context: Optional reading context.
        """
        if count == EOF:
            result = []
            while not _is_eof(stream):
                result.append(cls._read(stream, context))
            return result

        return [cls._read(stream, context) for _ in range(count)]

    def _read_0(cls, stream: BinaryIO, context: dict[str, Any] | None = None) -> list[Self]:
        """Internal function for reading null-terminated data.

        "Null" is type specific, so must be implemented per type.

        Args:
            stream: The stream to read from.
            context: Optional reading context.
        """
        raise NotImplementedError

    def _write(cls, stream: BinaryIO, data: Any) -> int:
        raise NotImplementedError

    def _write_array(cls, stream: BinaryIO, array: list[Self]) -> int:  # type: ignore
        """Internal function for writing arrays.

        Allows type implementations to do optimized writing for their type.

        Args:
            stream: The stream to read from.
            array: The array to write.
        """
        return sum(cls._write(stream, entry) for entry in array)

    def _write_0(cls, stream: BinaryIO, array: list[Self]) -> int:  # type: ignore
        """Internal function for writing null-terminated arrays.

        Allows type implementations to do optimized writing for their type.

        Args:
            stream: The stream to read from.
            array: The array to write.
        """
        return cls._write_array(stream, [*array, cls.__default__()])


class _overload:
    """Descriptor to use on the ``write`` and ``dumps`` methods on cstruct types.

    Allows for calling these methods on both the type and instance.

    Example:
        >>> int32.dumps(123)
        b'\\x7b\\x00\\x00\\x00'
        >>> int32(123).dumps()
        b'\\x7b\\x00\\x00\\x00'
    """

    def __init__(self, func: Callable[..., Any]) -> None:
        self.func = func

    def __get__(self, instance: BaseType | None, owner: type[BaseType]) -> Callable[[], bytes]:
        if instance is None:
            return functools.partial(self.func, owner)
        return functools.partial(self.func, instance.__class__, value=instance)


class BaseType(metaclass=MetaType):
    """Base class for cstruct type classes."""

    dumps = _overload(MetaType.dumps)
    write = _overload(MetaType.write)

    def __len__(self) -> int:
        """Return the byte size of the type."""
        if self.__class__.size is None:
            raise TypeError("Dynamic size")

        return self.__class__.size


T = TypeVar("T", bound=BaseType)


class BaseArray(BaseType):
    """Implements a fixed or dynamically sized array type.

    Example:
        When using the default C-style parser, the following syntax is supported:

            x[3] -> 3 -> static length.
            x[] -> None -> null-terminated.
            x[expr] -> expr -> dynamic length.
    """

    type: ClassVar[type[BaseType]]
    num_entries: ClassVar[int | Expression | None]
    null_terminated: ClassVar[bool]

    @classmethod
    def __default__(cls) -> BaseType:
        return type.__call__(
            cls, [cls.type.__default__()] * (cls.num_entries if isinstance(cls.num_entries, int) else 0)
        )

    @classmethod
    def _read(cls, stream: BinaryIO, context: dict[str, Any] | None = None) -> list[BaseType]:
        if cls.null_terminated:
            return cls.type._read_0(stream, context)

        if isinstance(cls.num_entries, int):
            num = max(0, cls.num_entries)
        elif cls.num_entries is None:
            num = EOF
        elif isinstance(cls.num_entries, Expression):
            try:
                num = max(0, cls.num_entries.evaluate(cls.cs, context))
            except Exception:
                if cls.num_entries.expression != "EOF":
                    raise
                num = EOF

        return cls.type._read_array(stream, num, context)

    @classmethod
    def _write(cls, stream: BinaryIO, data: list[Any]) -> int:
        if cls.null_terminated:
            return cls.type._write_0(stream, data)

        if not cls.dynamic and cls.num_entries != (actual_size := len(data)):
            raise ArraySizeError(f"Expected static array size {cls.num_entries}, got {actual_size} instead.")

        return cls.type._write_array(stream, data)


class Array(list[T], BaseArray):
    @classmethod
    def _read(cls, stream: BinaryIO, context: dict[str, Any] | None = None) -> list[T]:
        return cls(super()._read(stream, context))


def _is_readable_type(value: object) -> bool:
    return hasattr(value, "read")


def _is_buffer_type(value: object) -> bool:
    return isinstance(value, (bytes, memoryview, bytearray))


def _is_eof(stream: BinaryIO) -> bool:
    """Check if the stream has reached EOF."""
    pos = stream.tell()
    stream.read(1)

    if stream.tell() == pos:
        return True

    stream.seek(pos)
    return False


# As mentioned in the BaseType class, we correctly set the type here
MetaType.ArrayType = Array



class Error(Exception):
    pass


class ParserError(Error):
    pass


class ResolveError(Error):
    pass


class NullPointerDereference(Error):
    pass


class ArraySizeError(Error):
    pass


class ExpressionParserError(Error):
    pass


class ExpressionTokenizerError(Error):
    pass



from __future__ import annotations

import io
from collections import ChainMap
from collections.abc import Callable, MutableMapping
from contextlib import contextmanager
from enum import Enum
from functools import lru_cache
from itertools import chain
from operator import attrgetter
from textwrap import dedent
from types import FunctionType
from typing import TYPE_CHECKING, Any, BinaryIO

from dissect.cstruct.bitbuffer import BitBuffer
from dissect.cstruct.types.base import (
    BaseType,
    MetaType,
    _is_buffer_type,
    _is_readable_type,
)
from dissect.cstruct.types.enum import EnumMetaType
from dissect.cstruct.types.pointer import Pointer

if TYPE_CHECKING:
    from collections.abc import Iterator, Mapping
    from types import FunctionType

    from typing_extensions import Self


class Field:
    """Structure field."""

    def __init__(self, name: str | None, type_: type[BaseType], bits: int | None = None, offset: int | None = None):
        self.name = name  # The name of the field, or None if anonymous
        self._name = name or type_.__name__  # The name of the field, or the type name if anonymous
        self.type = type_
        self.bits = bits
        self.offset = offset
        self.alignment = type_.alignment or 1

    def __repr__(self) -> str:
        bits_str = f" : {self.bits}" if self.bits else ""
        return f"<Field {self.name} {self.type.__name__}{bits_str}>"


class StructureMetaType(MetaType):
    """Base metaclass for cstruct structure type classes."""

    # TODO: resolve field types in _update_fields, remove resolves elsewhere?

    fields: dict[str, Field]
    """Mapping of field names to :class:`Field` objects, including "folded" fields from anonymous structures."""
    lookup: dict[str, Field]
    """Mapping of "raw" field names to :class:`Field` objects. E.g. holds the anonymous struct and not its fields."""
    __fields__: list[Field]
    """List of :class:`Field` objects for this structure. This is the structures' Single Source Of Truth."""

    # Internal
    __align__: bool
    __anonymous__: bool
    __updating__ = False
    __compiled__ = False
    __static_sizes__: dict[str, int]  # Cache of static sizes by field name

    def __new__(metacls, name: str, bases: tuple[type, ...], classdict: dict[str, Any]) -> Self:  # type: ignore
        if (fields := classdict.pop("fields", None)) is not None:
            metacls._update_fields(metacls, fields, align=classdict.get("__align__", False), classdict=classdict)

        return super().__new__(metacls, name, bases, classdict)

    def __call__(cls, *args, **kwargs) -> Self:  # type: ignore
        if (
            cls.__fields__
            and len(args) == len(cls.__fields__) == 1
            and isinstance(args[0], bytes)
            and issubclass(cls.__fields__[0].type, bytes)
            and len(args[0]) == cls.__fields__[0].type.size
        ):
            # Shortcut for single char/bytes type
            return type.__call__(cls, *args, **kwargs)
        if not args and not kwargs:
            obj = type.__call__(cls)
            object.__setattr__(obj, "__dynamic_sizes__", {})
            return obj

        return super().__call__(*args, **kwargs)

    def _update_fields(
        cls, fields: list[Field], align: bool = False, classdict: dict[str, Any] | None = None
    ) -> dict[str, Any]:
        classdict = classdict or {}

        lookup = {}
        raw_lookup = {}
        field_names = []
        static_sizes = {}
        for field in fields:
            if field._name in lookup and field._name != "_":
                raise ValueError(f"Duplicate field name: {field._name}")

            if not field.type.dynamic:
                static_sizes[field._name] = field.type.size

            if isinstance(field.type, StructureMetaType) and field.name is None:
                for anon_field in field.type.fields.values():
                    attr = f"{field._name}.{anon_field.name}"
                    classdict[anon_field.name] = property(attrgetter(attr), attrsetter(attr))

                lookup.update(field.type.fields)
            else:
                lookup[field._name] = field

            raw_lookup[field._name] = field

        field_names = lookup.keys()
        classdict["fields"] = lookup
        classdict["lookup"] = raw_lookup
        classdict["__fields__"] = fields
        classdict["__static_sizes__"] = static_sizes
        classdict["__bool__"] = _generate__bool__(field_names)

        if issubclass(cls, UnionMetaType) or isinstance(cls, UnionMetaType):
            classdict["__init__"] = _generate_union__init__(raw_lookup.values())
            # Not a great way to do this but it works for now
            classdict["__eq__"] = Union.__eq__
        else:
            classdict["__init__"] = _generate_structure__init__(raw_lookup.values())
            classdict["__eq__"] = _generate__eq__(field_names)

        classdict["__hash__"] = _generate__hash__(field_names)

        # If we're calling this as a class method or a function on the metaclass
        if issubclass(cls, type):
            size, alignment = cls._calculate_size_and_offsets(cls, fields, align)
        else:
            size, alignment = cls._calculate_size_and_offsets(fields, align)

        if cls.__compiled__:
            # If the previous class was compiled try to compile this too
            from dissect.cstruct import compiler  # noqa: PLC0415

            try:
                classdict["_read"] = compiler.Compiler(cls.cs).compile_read(fields, cls.__name__, align=cls.__align__)
                classdict["__compiled__"] = True
            except Exception:
                # Revert _read to the slower loop based method
                classdict["_read"] = classmethod(Structure._read.__func__)
                classdict["__compiled__"] = False

        # TODO: compile _write
        # TODO: generate cached_property for lazy reading

        classdict["size"] = size
        classdict["alignment"] = alignment
        classdict["dynamic"] = size is None

        return classdict

    def _calculate_size_and_offsets(cls, fields: list[Field], align: bool = False) -> tuple[int | None, int]:
        """Iterate all fields in this structure to calculate the field offsets and total structure size.

        If a structure has a dynamic field, further field offsets will be set to None and self.dynamic
        will be set to True.
        """
        # The current offset, set to None if we become dynamic
        offset = 0
        # The current alignment for this structure
        alignment = 0

        # The current bit field type
        bits_type = None
        # The offset of the current bit field, set to None if we become dynamic
        bits_field_offset = 0
        # How many bits we have left in the current bit field
        bits_remaining = 0

        for field in fields:
            if field.offset is not None:
                # If a field already has an offset, it's leading
                offset = field.offset

            if align and offset is not None:
                # Round to next alignment
                offset += -offset & (field.alignment - 1)

            # The alignment of this struct is equal to its largest members' alignment
            alignment = max(alignment, field.alignment)

            if field.bits:
                field_type = field.type

                if isinstance(field_type, EnumMetaType):
                    field_type = field_type.type

                # Bit fields have special logic
                if (
                    # Exhausted a bit field
                    bits_remaining == 0
                    # Moved to a bit field of another type, e.g. uint16 f1 : 8, uint32 f2 : 8;
                    or field_type != bits_type
                    # Still processing a bit field, but it's at a different offset due to alignment or a manual offset
                    or (bits_type is not None and offset > bits_field_offset + bits_type.size)
                ):
                    # ... if any of this is true, we have to move to the next field
                    bits_type = field_type
                    bits_count = bits_type.size * 8
                    bits_remaining = bits_count
                    bits_field_offset = offset

                    if offset is not None:
                        # We're not dynamic, update the structure size and current offset
                        offset += bits_type.size

                    field.offset = bits_field_offset

                bits_remaining -= field.bits

                if bits_remaining < 0:
                    raise ValueError("Straddled bit fields are unsupported")
            else:
                # Reset bits stuff
                bits_type = None
                bits_field_offset = bits_remaining = 0

                field.offset = offset

                if offset is not None:
                    # We're not dynamic, update the structure size and current offset
                    try:
                        field_len = len(field.type)
                    except TypeError:
                        # This field is dynamic
                        offset = None
                        continue

                    offset += field_len

        if align and offset is not None:
            # Add "tail padding" if we need to align
            # This bit magic rounds up to the next alignment boundary
            # E.g. offset = 3; alignment = 8; -offset & (alignment - 1) = 5
            offset += -offset & (alignment - 1)

        # The structure size is whatever the currently calculated offset is
        return offset, alignment

    def _read(cls, stream: BinaryIO, context: dict[str, Any] | None = None) -> Self:  # type: ignore
        bit_buffer = BitBuffer(stream, cls.cs.endian)
        struct_start = stream.tell()

        result = {}
        sizes = {}
        for field in cls.__fields__:
            offset = stream.tell()

            if field.offset is not None and offset != struct_start + field.offset:
                # Field is at a specific offset, either alligned or added that way
                offset = struct_start + field.offset
                stream.seek(offset)

            if cls.__align__ and field.offset is None:
                # Previous field was dynamically sized and we need to align
                offset += -offset & (field.alignment - 1)
                stream.seek(offset)

            if field.bits:
                if isinstance(field.type, EnumMetaType):
                    value = field.type(bit_buffer.read(field.type.type, field.bits))
                else:
                    value = bit_buffer.read(field.type, field.bits)

                result[field._name] = value
                continue

            bit_buffer.reset()

            value = field.type._read(stream, result)

            result[field._name] = value
            if field.type.dynamic:
                sizes[field._name] = stream.tell() - offset

        if cls.__align__:
            # Align the stream
            stream.seek(-stream.tell() & (cls.alignment - 1), io.SEEK_CUR)

        # Using type.__call__ directly calls the __init__ method of the class
        # This is faster than calling cls() and bypasses the metaclass __call__ method
        obj = type.__call__(cls, **result)
        obj.__dynamic_sizes__ = sizes
        return obj

    def _read_0(cls, stream: BinaryIO, context: dict[str, Any] | None = None) -> list[Self]:  # type: ignore
        result = []

        while obj := cls._read(stream, context):
            result.append(obj)

        return result

    def _write(cls, stream: BinaryIO, data: Structure) -> int:
        bit_buffer = BitBuffer(stream, cls.cs.endian)
        struct_start = stream.tell()
        num = 0

        for field in cls.__fields__:
            field_type = cls.cs.resolve(field.type)

            bit_field_type = (
                (field_type.type if isinstance(field_type, EnumMetaType) else field_type) if field.bits else None
            )
            # Current field is not a bit field, but previous was
            # Or, moved to a bit field of another type, e.g. uint16 f1 : 8, uint32 f2 : 8;
            if (not field.bits and bit_buffer._type is not None) or (
                bit_buffer._type and bit_buffer._type != bit_field_type
            ):
                # Flush the current bit buffer so we can process alignment properly
                bit_buffer.flush()

            offset = stream.tell()

            if field.offset is not None and offset < struct_start + field.offset:
                # Field is at a specific offset, either alligned or added that way
                stream.write(b"\x00" * (struct_start + field.offset - offset))
                offset = struct_start + field.offset

            if cls.__align__ and field.offset is None:
                is_bitbuffer_boundary = bit_buffer._type and (
                    bit_buffer._remaining == 0 or bit_buffer._type != field_type
                )
                if not bit_buffer._type or is_bitbuffer_boundary:
                    # Previous field was dynamically sized and we need to align
                    align_pad = -offset & (field.alignment - 1)
                    stream.write(b"\x00" * align_pad)
                    offset += align_pad

            value = getattr(data, field._name, None)
            if value is None:
                value = field_type.__default__()

            if field.bits:
                if isinstance(field_type, EnumMetaType):
                    bit_buffer.write(field_type.type, value.value, field.bits)
                else:
                    bit_buffer.write(field_type, value, field.bits)
            else:
                field_type._write(stream, value)
                num += stream.tell() - offset

        if bit_buffer._type is not None:
            bit_buffer.flush()

        if cls.__align__:
            # Align the stream
            stream.write(b"\x00" * (-stream.tell() & (cls.alignment - 1)))

        return num

    def add_field(cls, name: str, type_: type[BaseType], bits: int | None = None, offset: int | None = None) -> None:
        field = Field(name, type_, bits=bits, offset=offset)
        cls.__fields__.append(field)

        if not cls.__updating__:
            cls.commit()

    @contextmanager
    def start_update(cls) -> Iterator[None]:
        try:
            cls.__updating__ = True
            yield
        finally:
            cls.commit()
            cls.__updating__ = False

    def commit(cls) -> None:
        classdict = cls._update_fields(cls.__fields__, cls.__align__)

        for key, value in classdict.items():
            setattr(cls, key, value)


class Structure(BaseType, metaclass=StructureMetaType):
    """Base class for cstruct structure type classes.

    Note that setting attributes which do not correspond to a field in the structure results in undefined behavior.
    For performance reasons, the structure does not check if the field exists when writing to an attribute.
    """

    __dynamic_sizes__: dict[str, int]

    def __len__(self) -> int:
        return len(self.dumps())

    def __bytes__(self) -> bytes:
        return self.dumps()

    def __getitem__(self, item: str) -> Any:
        return getattr(self, item)

    def __repr__(self) -> str:
        values = []
        for name, field in self.__class__.fields.items():
            value = self[name]
            if issubclass(field.type, int) and not issubclass(field.type, (Pointer, Enum)):
                value = hex(value)
            else:
                value = repr(value)
            values.append(f"{name}={value}")

        return f"<{self.__class__.__name__} {' '.join(values)}>"

    @property
    def __values__(self) -> MutableMapping[str, Any]:
        return StructureValuesProxy(self)

    @property
    def __sizes__(self) -> Mapping[str, int | None]:
        return ChainMap(self.__class__.__static_sizes__, self.__dynamic_sizes__)


class StructureValuesProxy(MutableMapping):
    """A proxy for the values of fields of a Structure."""

    def __init__(self, struct: Structure):
        self._struct: Structure = struct

    def __getitem__(self, key: str) -> Any:
        if key in self:
            return getattr(self._struct, key)
        raise KeyError(key)

    def __setitem__(self, key: str, value: Any) -> None:
        if key in self:
            return setattr(self._struct, key, value)
        raise KeyError(key)

    def __contains__(self, key: str) -> bool:
        return key in self._struct.__class__.fields

    def __iter__(self) -> Iterator[str]:
        return iter(self._struct.__class__.fields)

    def __len__(self) -> int:
        return len(self._struct.__class__.fields)

    def __repr__(self) -> str:
        return repr(dict(self))

    def __delitem__(self, _: str):
        # Is abstract in base, but deleting is not supported.
        raise NotImplementedError("Cannot delete fields from a Structure")


class UnionMetaType(StructureMetaType):
    """Base metaclass for cstruct union type classes."""

    def __call__(cls, *args, **kwargs) -> Self:  # type: ignore
        obj: Union = super().__call__(*args, **kwargs)

        # Calling with non-stream args or kwargs means we are initializing with values
        if (args and not (len(args) == 1 and (_is_readable_type(args[0]) or _is_buffer_type(args[0])))) or kwargs:
            # We don't support user initialization of dynamic unions yet
            if cls.dynamic:
                raise NotImplementedError("Initializing a dynamic union is not yet supported")

            # User (partial) initialization, rebuild the union
            # First user-provided field is the one used to rebuild the union
            arg_fields = (field._name for _, field in zip(args, cls.__fields__, strict=False))
            kwarg_fields = (name for name in kwargs if name in cls.lookup)
            if (first_field := next(chain(arg_fields, kwarg_fields), None)) is not None:
                obj._rebuild(first_field)
        elif not args and not kwargs:
            # Initialized with default values
            # Note that we proxify here in case we have a default initialization (cls())
            # We don't proxify in case we read from a stream, as we do that later on in _read at a more appropriate time
            # Same with (partial) user initialization, we do that after rebuilding the union
            obj._proxify()

        return obj

    def _calculate_size_and_offsets(cls, fields: list[Field], align: bool = False) -> tuple[int | None, int]:
        size = 0
        alignment = 0

        for field in fields:
            if size is not None:
                try:
                    size = max(len(field.type), size)
                except TypeError:
                    size = None

            alignment = max(field.alignment, alignment)

        if align and size is not None:
            # Add "tail padding" if we need to align
            # This bit magic rounds up to the next alignment boundary
            # E.g. offset = 3; alignment = 8; -offset & (alignment - 1) = 5
            size += -size & (alignment - 1)

        return size, alignment

    def _read_fields(
        cls, stream: BinaryIO, context: dict[str, Any] | None = None
    ) -> tuple[dict[str, Any], dict[str, int]]:
        result = {}
        sizes = {}

        if cls.size is None:
            offset = stream.tell()
            buf = stream
        else:
            offset = 0
            buf = io.BytesIO(stream.read(cls.size))

        for field in cls.__fields__:
            field_type = cls.cs.resolve(field.type)

            start = 0
            if field.offset is not None:
                start = field.offset

            buf.seek(offset + start)
            value = field_type._read(buf, result)

            result[field._name] = value
            if field.type.dynamic:
                sizes[field._name] = buf.tell() - start

        return result, sizes

    def _read(cls, stream: BinaryIO, context: dict[str, Any] | None = None) -> Self:  # type: ignore
        if cls.size is None:
            start = stream.tell()
            result, sizes = cls._read_fields(stream, context)
            size = stream.tell() - start
            stream.seek(start)
            buf = stream.read(size)
        else:
            result = {}
            sizes = {}
            buf = stream.read(cls.size)

        # Create the object and set the values
        # Using type.__call__ directly calls the __init__ method of the class
        # This is faster than calling cls() and bypasses the metaclass __call__ method
        # It also makes it easier to differentiate between user-initialization of the class
        # and initialization from a stream read
        obj: Union = type.__call__(cls, **result)
        object.__setattr__(obj, "__dynamic_sizes__", sizes)
        object.__setattr__(obj, "_buf", buf)

        if cls.size is not None:
            obj._update()

        # Proxify any nested structures
        obj._proxify()

        return obj

    def _write(cls, stream: BinaryIO, data: Union) -> int:
        if cls.dynamic:
            raise NotImplementedError("Writing dynamic unions is not yet supported")

        offset = stream.tell()
        expected_offset = offset + len(cls)

        # Sort by largest field
        fields = sorted(cls.__fields__, key=lambda e: e.type.size or 0, reverse=True)
        anonymous_struct = False

        # Try to write by largest field
        for field in fields:
            if isinstance(field.type, StructureMetaType) and field.name is None:
                # Prefer to write regular fields initially
                anonymous_struct = field.type
                continue

            # Write the value
            field.type._write(stream, getattr(data, field._name))
            break

        # If we haven't written anything yet and we initially skipped an anonymous struct, write it now
        if stream.tell() == offset and anonymous_struct:
            anonymous_struct._write(stream, data)

        # If we haven't filled the union size yet, pad it
        if remaining := expected_offset - stream.tell():
            stream.write(b"\x00" * remaining)

        return stream.tell() - offset


class Union(Structure, metaclass=UnionMetaType):
    """Base class for cstruct union type classes."""

    _buf: bytes

    def __eq__(self, other: object) -> bool:
        return self.__class__ is other.__class__ and bytes(self) == bytes(other)

    def __setattr__(self, attr: str, value: Any) -> None:
        if self.__class__.dynamic:
            raise NotImplementedError("Modifying a dynamic union is not yet supported")

        super().__setattr__(attr, value)
        self._rebuild(attr)

    def _rebuild(self, attr: str) -> None:
        if (cur_buf := getattr(self, "_buf", None)) is None:
            cur_buf = b"\x00" * self.__class__.size

        buf = io.BytesIO(cur_buf)
        field = self.__class__.lookup[attr]
        if field.offset:
            buf.seek(field.offset)

        if (value := getattr(self, attr)) is None:
            value = field.type.__default__()

        field.type._write(buf, value)

        object.__setattr__(self, "_buf", buf.getvalue())
        self._update()

        # (Re-)proxify all values
        self._proxify()

    def _update(self) -> None:
        result, sizes = self.__class__._read_fields(io.BytesIO(self._buf))
        self.__dict__.update(result)
        object.__setattr__(self, "__dynamic_sizes__", sizes)

    def _proxify(self) -> None:
        def _proxy_structure(value: Structure) -> None:
            for field in value.__class__.__fields__:
                if issubclass(field.type, Structure):
                    nested_value = getattr(value, field._name)
                    proxy = UnionProxy(self, field._name, nested_value)
                    object.__setattr__(value, field._name, proxy)
                    _proxy_structure(nested_value)

        _proxy_structure(self)


class UnionProxy:
    __union__: Union
    __attr__: str
    __target__: Structure

    def __init__(self, union: Union, attr: str, target: Structure):
        object.__setattr__(self, "__union__", union)
        object.__setattr__(self, "__attr__", attr)
        object.__setattr__(self, "__target__", target)

    def __len__(self) -> int:
        return len(self.__target__.dumps())

    def __bytes__(self) -> bytes:
        return self.__target__.dumps()

    def __getitem__(self, item: str) -> Any:
        return getattr(self.__target__, item)

    def __repr__(self) -> str:
        return repr(self.__target__)

    def __getattr__(self, attr: str) -> Any:
        return getattr(self.__target__, attr)

    def __setattr__(self, attr: str, value: Any) -> None:
        setattr(self.__target__, attr, value)
        self.__union__._rebuild(self.__attr__)


def attrsetter(path: str) -> Callable[[Any], Any]:
    path, _, attr = path.rpartition(".")
    path = path.split(".")

    def _func(obj: Any, value: Any) -> Any:
        for name in path:
            obj = getattr(obj, name)
        setattr(obj, attr, value)

    return _func


def _codegen(func: FunctionType) -> FunctionType:
    """Decorator that generates a template function with a specified number of fields.

    This code is a little complex but allows use to cache generated functions for a specific number of fields.
    For example, if we generate a structure with 10 fields, we can cache the generated code for that structure.
    We can then reuse that code and patch it with the correct field names when we create a new structure with 10 fields.

    The functions that are decorated with this decorator should take a list of field names and return a string of code.
    The decorated function is needs to be called with the number of fields, instead of the field names.
    The confusing part is that that the original function takes field names, but you then call it with
    the number of fields instead.

    Inspired by https://github.com/dabeaz/dataklasses.

    Args:
        func: The decorated function that takes a list of field names and returns a string of code.

    Returns:
        A cached function that generates the desired function code, to be called with the number of fields.
    """

    def make_func_code(num_fields: int) -> FunctionType:
        exec(func([f"_{n}" for n in range(num_fields)]), {}, d := {})
        return d.popitem()[1]

    make_func_code.__wrapped__ = func
    return lru_cache(make_func_code)


@_codegen
def _make_structure__init__(fields: list[str]) -> str:
    """Generates an ``__init__`` method for a structure with the specified fields.

    Args:
        fields: List of field names.
    """
    field_args = ", ".join(f"{field} = None" for field in fields)
    field_init = "\n".join(
        f" self.{name} = {name} if {name} is not None else _{i}_default" for i, name in enumerate(fields)
    )

    code = f"def __init__(self{', ' + field_args or ''}):\n"
    return code + (field_init or " pass")


@_codegen
def _make_union__init__(fields: list[str]) -> str:
    """Generates an ``__init__`` method for a class with the specified fields using setattr.

    Args:
        fields: List of field names.
    """
    field_args = ", ".join(f"{field} = None" for field in fields)
    field_init = "\n".join(
        f" object.__setattr__(self, '{name}', {name} if {name} is not None else _{i}_default)"
        for i, name in enumerate(fields)
    )

    code = f"def __init__(self{', ' + field_args or ''}):\n"
    return code + (field_init or " pass")


@_codegen
def _make__eq__(fields: list[str]) -> str:
    """Generates an ``__eq__`` method for a class with the specified fields.

    Args:
        fields: List of field names.
    """
    self_vals = ",".join(f"self.{name}" for name in fields)
    other_vals = ",".join(f"other.{name}" for name in fields)

    if self_vals:
        self_vals += ","
    if other_vals:
        other_vals += ","

    # In the future this could be a looser check, e.g. an __eq__ on the classes, which compares the fields
    code = f"""
    def __eq__(self, other):
        if self.__class__ is other.__class__:
            return ({self_vals}) == ({other_vals})
        return False
    """

    return dedent(code)


@_codegen
def _make__bool__(fields: list[str]) -> str:
    """Generates a ``__bool__`` method for a class with the specified fields.

    Args:
        fields: List of field names.
    """
    vals = ", ".join(f"self.{name}" for name in fields)

    code = f"""
    def __bool__(self):
        return any([{vals}])
    """

    return dedent(code)


@_codegen
def _make__hash__(fields: list[str]) -> str:
    """Generates a ``__hash__`` method for a class with the specified fields.

    Args:
        fields: List of field names.
    """
    vals = ", ".join(f"self.{name}" for name in fields)

    code = f"""
    def __hash__(self):
        return hash(({vals}))
    """

    return dedent(code)


def _patch_attributes(func: FunctionType, fields: list[str], start: int = 0) -> FunctionType:
    """Patches a function's attributes.

    Args:
        func: The function to patch.
        fields: List of field names to add.
        start: The starting index for patching. Defaults to 0.
    """
    return type(func)(
        func.__code__.replace(co_names=(*func.__code__.co_names[:start], *fields)),
        func.__globals__,
    )


def _generate_structure__init__(fields: list[Field]) -> FunctionType:
    """Generates an ``__init__`` method for a structure with the specified fields.

    Args:
        fields: List of field names.
    """
    mapping = _generate_co_mapping(fields)

    template: FunctionType = _make_structure__init__(len(fields))
    return type(template)(
        template.__code__.replace(
            co_names=_remap_co_values(template.__code__.co_names, mapping),
            co_varnames=_remap_co_values(template.__code__.co_varnames, mapping),
        ),
        template.__globals__ | {f"__{field._name}_default__": field.type.__default__() for field in fields},
        argdefs=template.__defaults__,
    )


def _generate_union__init__(fields: list[Field]) -> FunctionType:
    """Generates an ``__init__`` method for a union with the specified fields.

    Args:
        fields: List of field names.
    """
    mapping = _generate_co_mapping(fields)

    template: FunctionType = _make_union__init__(len(fields))
    return type(template)(
        template.__code__.replace(
            co_consts=_remap_co_values(template.__code__.co_consts, mapping),
            co_names=_remap_co_values(template.__code__.co_names, mapping),
            co_varnames=_remap_co_values(template.__code__.co_varnames, mapping),
        ),
        template.__globals__ | {f"__{field._name}_default__": field.type.__default__() for field in fields},
        argdefs=template.__defaults__,
    )


def _generate_co_mapping(fields: list[Field]) -> dict[str, str]:
    """Generates a mapping of generated code object names to field names.

    The generated code uses names like ``_0``, ``_1``, etc. for fields, and ``_0_default``, ``_1_default``, etc.
    for default initializer values. Return a mapping of these names to the actual field names.

    Args:
        fields: List of field names.
    """
    return {
        key: value
        for i, field in enumerate(fields)
        for key, value in [(f"_{i}", field._name), (f"_{i}_default", f"__{field._name}_default__")]
    }


def _remap_co_values(value: tuple[Any, ...], mapping: dict[str, str]) -> tuple[Any, ...]:
    """Remaps code object values using a mapping.

    This is used to replace generated code object names with actual field names.

    Args:
        value: The original code object values.
        mapping: A mapping of generated code object names to field names.
    """
    # Only attempt to remap if the value is a string, otherwise return it as is
    # This is to avoid issues with trying to remap non-hashable types, and we only need to replace strings anyway
    return tuple(mapping.get(v, v) if isinstance(v, str) else v for v in value)


def _generate__eq__(fields: list[str]) -> FunctionType:
    """Generates an ``__eq__`` method for a class with the specified fields.

    Args:
        fields: List of field names.
    """
    return _patch_attributes(_make__eq__(len(fields)), fields, 1)


def _generate__bool__(fields: list[str]) -> FunctionType:
    """Generates a ``__bool__`` method for a class with the specified fields.

    Args:
        fields: List of field names.
    """
    return _patch_attributes(_make__bool__(len(fields)), fields, 1)


def _generate__hash__(fields: list[str]) -> FunctionType:
    """Generates a ``__hash__`` method for a class with the specified fields.

    Args:
        fields: List of field names.
    """
    return _patch_attributes(_make__hash__(len(fields)), fields, 1)



# Made in Japan

from __future__ import annotations

import io
import logging
from enum import Enum
from textwrap import dedent, indent
from typing import TYPE_CHECKING

from dissect.cstruct.bitbuffer import BitBuffer
from dissect.cstruct.types import (
    Array,
    BaseType,
    Char,
    CharArray,
    Flag,
    Int,
    Packed,
    Pointer,
    Structure,
    Union,
    Void,
    VoidArray,
    Wchar,
    WcharArray,
)
from dissect.cstruct.types.base import BaseArray
from dissect.cstruct.types.enum import EnumMetaType
from dissect.cstruct.types.packed import _struct

if TYPE_CHECKING:
    from collections.abc import Iterator
    from types import MethodType

    from dissect.cstruct.cstruct import cstruct
    from dissect.cstruct.types.structure import Field

SUPPORTED_TYPES = (
    Array,
    Char,
    CharArray,
    Enum,
    Flag,
    Int,
    Packed,
    Pointer,
    Structure,
    Void,
    Wchar,
    WcharArray,
    VoidArray,
)

log = logging.getLogger(__name__)

python_compile = compile


def compile(structure: type[Structure]) -> type[Structure]:
    return Compiler(structure.cs).compile(structure)


class Compiler:
    def __init__(self, cs: cstruct):
        self.cs = cs

    def compile(self, structure: type[Structure]) -> type[Structure]:
        if issubclass(structure, Union):
            return structure

        try:
            structure._read = self.compile_read(structure.__fields__, structure.__name__, structure.__align__)
            structure.__compiled__ = True
        except Exception as e:
            # Silently ignore, we didn't compile unfortunately
            log.debug("Failed to compile %s", structure, exc_info=e)

        return structure

    def compile_read(self, fields: list[Field], name: str | None = None, align: bool = False) -> MethodType:
        return _ReadSourceGenerator(self.cs, fields, name, align).generate()


class _ReadSourceGenerator:
    def __init__(self, cs: cstruct, fields: list[Field], name: str | None = None, align: bool = False):
        self.cs = cs
        self.fields = fields
        self.name = name
        self.align = align

        self.field_map: dict[str, Field] = {}
        self._token_id = 0

    def _map_field(self, field: Field) -> str:
        token = f"_{self._token_id}"
        self.field_map[token] = field
        self._token_id += 1
        return token

    def generate(self) -> MethodType:
        source = self.generate_source()
        symbols = {token: field.type for token, field in self.field_map.items()}

        code = python_compile(source, f"<compiled {self.name or 'anonymous'}._read>", "exec")
        exec(code, {"BitBuffer": BitBuffer, "_struct": _struct, **symbols}, d := {})
        obj = d.popitem()[1]
        obj.__source__ = source

        return classmethod(obj)

    def generate_source(self) -> str:
        preamble = """
        r = {}
        s = {}
        o = stream.tell()
        """

        if any(field.bits for field in self.fields):
            preamble += "bit_reader = BitBuffer(stream, cls.cs.endian)\n"

        read_code = "\n".join(self._generate_fields())

        outro = """
        obj = type.__call__(cls, **r)
        obj.__dynamic_sizes__ = s

        return obj
        """

        code = indent(dedent(preamble).lstrip() + read_code + dedent(outro), "    ")

        return f"def _read(cls, stream, context=None):\n{code}"

    def _generate_fields(self) -> Iterator[str]:
        current_offset = 0
        current_block: list[Field] = []
        prev_was_bits = False
        prev_bits_type = None
        bits_remaining = 0
        bits_rollover = False

        def flush() -> Iterator[str]:
            if current_block:
                if self.align and current_block[0].offset is None:
                    yield f"stream.seek(-stream.tell() & ({current_block[0].alignment} - 1), {io.SEEK_CUR})"

                yield from self._generate_packed(current_block)
                current_block[:] = []

        def align_to_field(field: Field) -> Iterator[str]:
            nonlocal current_offset

            if field.offset is not None and field.offset != current_offset:
                # If a field has a set offset and it's not the same as the current tracked offset, seek to it
                yield f"stream.seek(o + {field.offset})"
                current_offset = field.offset

            if self.align and field.offset is None:
                yield f"stream.seek(-stream.tell() & ({field.alignment} - 1), {io.SEEK_CUR})"

        for field in self.fields:
            field_type = field.type

            if isinstance(field_type, EnumMetaType):
                field_type = field_type.type

            if not issubclass(field_type, SUPPORTED_TYPES):
                raise TypeError(f"Unsupported type for compiler: {field_type}")

            if prev_was_bits and not field.bits:
                yield "bit_reader.reset()"
                prev_was_bits = False
                bits_remaining = 0

            try:
                size = len(field_type)
                is_dynamic = False
            except TypeError:
                size = None
                is_dynamic = True

            # Sub structure
            if issubclass(field_type, Structure):
                yield from flush()
                yield from align_to_field(field)
                yield from self._generate_structure(field)

            # Array of structures and multi-dimensional arrays
            elif issubclass(field_type, (Array, CharArray, WcharArray)) and (
                issubclass(field_type.type, Structure) or issubclass(field_type.type, BaseArray) or is_dynamic
            ):
                yield from flush()
                yield from align_to_field(field)
                yield from self._generate_array(field)

            # Bit fields
            elif field.bits:
                if size is None:
                    raise TypeError(f"Unsupported type for bit field: {field_type}")

                if not prev_was_bits:
                    prev_bits_type = field_type
                    prev_was_bits = True

                if bits_remaining == 0 or prev_bits_type != field_type:
                    bits_remaining = (size * 8) - field.bits
                    bits_rollover = True

                yield from flush()
                yield from align_to_field(field)
                yield from self._generate_bits(field)

            # Everything else - basic and composite types (and arrays of them)
            else:
                current_block.append(field)

            if current_offset is not None and size is not None and (not field.bits or bits_rollover):
                current_offset += size
                bits_rollover = False

        yield from flush()

        if self.align:
            yield f"stream.seek(-stream.tell() & (cls.alignment - 1), {io.SEEK_CUR})"

    def _generate_structure(self, field: Field) -> Iterator[str]:
        template = f"""
        {"_s = stream.tell()" if field.type.dynamic else ""}
        r["{field._name}"] = {self._map_field(field)}._read(stream, context=r)
        {f's["{field._name}"] = stream.tell() - _s' if field.type.dynamic else ""}
        """

        yield dedent(template)

    def _generate_array(self, field: Field) -> Iterator[str]:
        template = f"""
        {"_s = stream.tell()" if field.type.dynamic else ""}
        r["{field._name}"] = {self._map_field(field)}._read(stream, context=r)
        {f's["{field._name}"] = stream.tell() - _s' if field.type.dynamic else ""}
        """

        yield dedent(template)

    def _generate_bits(self, field: Field) -> Iterator[str]:
        lookup = self._map_field(field)
        read_type = "_t"
        field_type = field.type
        if issubclass(field_type, (Enum, Flag)):
            read_type += ".type"
            field_type = field_type.type

        if issubclass(field_type, Char):
            field_type = field_type.cs.uint8
            lookup = "cls.cs.uint8"

        template = f"""
        _t = {lookup}
        r["{field._name}"] = type.__call__(_t, bit_reader.read({read_type}, {field.bits}))
        """

        yield dedent(template)

    def _generate_packed(self, fields: list[Field]) -> Iterator[str]:
        info = list(_generate_struct_info(self.cs, fields, self.align))
        reads = []

        size = 0
        slice_index = 0
        for field, count, _ in info:
            if field is None:
                # Padding
                size += count
                continue

            field_type = field.type
            read_type = _get_read_type(self.cs, field_type)

            if issubclass(field_type, (Array, CharArray, WcharArray)):
                count = field_type.num_entries
                read_type = _get_read_type(self.cs, field_type.type)

                if issubclass(read_type, (Char, Wchar, Int)):
                    count *= read_type.size
                    getter = f"buf[{size}:{size + count}]"
                else:
                    getter = f"data[{slice_index}:{slice_index + count}]"
                    slice_index += count
            elif issubclass(read_type, (Char, Wchar, Int)):
                getter = f"buf[{size}:{size + read_type.size}]"
            else:
                getter = f"data[{slice_index}]"
                slice_index += 1

            if issubclass(read_type, (Wchar, Int)):
                # Types that parse bytes further down to their own type
                parser_template = "{type}({getter})"
            else:
                # All other types can be simply intialized
                parser_template = "type.__call__({type}, {getter})"

            # Create the final reading code
            if issubclass(field_type, Array):
                reads.append(f"_t = {self._map_field(field)}")
                reads.append("_et = _t.type")

                if issubclass(field_type.type, Int):
                    reads.append(f"_b = {getter}")
                    item_parser = parser_template.format(type="_et", getter=f"_b[i:i + {field_type.type.size}]")
                    list_comp = f"[{item_parser} for i in range(0, {count}, {field_type.type.size})]"
                elif issubclass(field_type.type, Pointer):
                    item_parser = "_et.__new__(_et, e, stream, r)"
                    list_comp = f"[{item_parser} for e in {getter}]"
                else:
                    item_parser = parser_template.format(type="_et", getter="e")
                    list_comp = f"[{item_parser} for e in {getter}]"

                parser = f"type.__call__(_t, {list_comp})"
            elif issubclass(field_type, CharArray):
                parser = f"type.__call__({self._map_field(field)}, {getter})"
            elif issubclass(field_type, Pointer):
                reads.append(f"_pt = {self._map_field(field)}")
                parser = f"_pt.__new__(_pt, {getter}, stream, r)"
            else:
                parser = parser_template.format(type=self._map_field(field), getter=getter)

            reads.append(f'r["{field._name}"] = {parser}')
            reads.append("")  # Generates a newline in the resulting code

            size += field_type.size

        fmt = _optimize_struct_fmt(info)
        if fmt == "x" or (len(fmt) == 2 and fmt[1] == "x"):
            unpack = ""
        else:
            unpack = f'data = _struct(cls.cs.endian, "{fmt}").unpack(buf)\n'

        template = f"""
        buf = stream.read({size})
        if len(buf) != {size}: raise EOFError()
        {unpack}
        """

        yield dedent(template) + "\n".join(reads)


def _generate_struct_info(cs: cstruct, fields: list[Field], align: bool = False) -> Iterator[tuple[Field, int, str]]:
    if not fields:
        return

    current_offset = fields[0].offset
    imaginary_offset = 0
    for field in fields:
        # We moved -- probably due to alignment
        if field.offset is not None and (drift := field.offset - current_offset) > 0:
            yield None, drift, "x"
            current_offset += drift

        if align and field.offset is None and (drift := -imaginary_offset & (field.alignment - 1)) > 0:
            # Assume we started at a correctly aligned boundary
            yield None, drift, "x"
            imaginary_offset += drift

        count = 1
        read_type = _get_read_type(cs, field.type)

        # Drop voids
        if issubclass(read_type, (Void, VoidArray)):
            continue

        # Array of more complex types are handled elsewhere
        if issubclass(read_type, (Array, CharArray, WcharArray)):
            count = read_type.num_entries
            read_type = _get_read_type(cs, read_type.type)

        # Take the pack char for Packed
        if issubclass(read_type, Packed):
            yield field, count, read_type.packchar

        # Other types are byte based
        # We don't actually unpack anything here but slice directly out of the buffer
        elif issubclass(read_type, (Char, Wchar, Int)):
            yield field, count * read_type.size, "x"

        size = count * read_type.size
        imaginary_offset += size
        if current_offset is not None:
            current_offset += size


def _optimize_struct_fmt(info: Iterator[tuple[Field, int, str]]) -> str:
    chars = []

    current_count = 0
    current_char = None

    for _, count, char in info:
        if current_char is None:
            current_count = count
            current_char = char
            continue

        if char != current_char:
            if current_count:
                chars.append((current_count, current_char))
            current_count = count
            current_char = char
        else:
            current_count += count

    if current_char is not None and current_count:
        chars.append((current_count, current_char))

    return "".join(f"{count if count > 1 else ''}{char}" for count, char in chars)


def _get_read_type(cs: cstruct, type_: type[BaseType]) -> type[BaseType]:
    if issubclass(type_, (Enum, Flag)):
        type_ = type_.type

    if issubclass(type_, Pointer):
        type_ = cs.pointer

    return cs.resolve(type_)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, BinaryIO, Generic, TypeVar

from dissect.cstruct.exceptions import NullPointerDereference
from dissect.cstruct.types.base import BaseType
from dissect.cstruct.types.char import Char
from dissect.cstruct.types.void import Void

if TYPE_CHECKING:
    from typing_extensions import Self

T = TypeVar("T", bound=BaseType)


class Pointer(int, BaseType, Generic[T]):
    """Pointer to some other type."""

    type: type[T]
    _stream: BinaryIO | None
    _context: dict[str, Any] | None
    _value: T | None

    def __new__(cls, value: int, stream: BinaryIO | None, context: dict[str, Any] | None = None) -> Self:
        obj = super().__new__(cls, value)
        obj._stream = stream
        obj._context = context
        obj._value = None
        return obj

    def __repr__(self) -> str:
        return f"<{self.type.__name__}* @ {self:#x}>"

    def __str__(self) -> str:
        return str(self.dereference())

    def __getattr__(self, attr: str) -> Any:
        return getattr(self.dereference(), attr)

    def __add__(self, other: int) -> Self:
        return type.__call__(self.__class__, int.__add__(self, other), self._stream, self._context)

    def __sub__(self, other: int) -> Self:
        return type.__call__(self.__class__, int.__sub__(self, other), self._stream, self._context)

    def __mul__(self, other: int) -> Self:
        return type.__call__(self.__class__, int.__mul__(self, other), self._stream, self._context)

    def __floordiv__(self, other: int) -> Self:
        return type.__call__(self.__class__, int.__floordiv__(self, other), self._stream, self._context)

    def __mod__(self, other: int) -> Self:
        return type.__call__(self.__class__, int.__mod__(self, other), self._stream, self._context)

    def __pow__(self, other: int) -> Self:
        return type.__call__(self.__class__, int.__pow__(self, other), self._stream, self._context)

    def __lshift__(self, other: int) -> Self:
        return type.__call__(self.__class__, int.__lshift__(self, other), self._stream, self._context)

    def __rshift__(self, other: int) -> Self:
        return type.__call__(self.__class__, int.__rshift__(self, other), self._stream, self._context)

    def __and__(self, other: int) -> Self:
        return type.__call__(self.__class__, int.__and__(self, other), self._stream, self._context)

    def __xor__(self, other: int) -> Self:
        return type.__call__(self.__class__, int.__xor__(self, other), self._stream, self._context)

    def __or__(self, other: int) -> Self:
        return type.__call__(self.__class__, int.__or__(self, other), self._stream, self._context)

    @classmethod
    def __default__(cls) -> Self:
        return cls.__new__(cls, cls.cs.pointer.__default__(), None, None)

    @classmethod
    def _read(cls, stream: BinaryIO, context: dict[str, Any] | None = None) -> Self:
        return cls.__new__(cls, cls.cs.pointer._read(stream, context), stream, context)

    @classmethod
    def _write(cls, stream: BinaryIO, data: int) -> int:
        return cls.cs.pointer._write(stream, data)

    def dereference(self) -> T:
        if self == 0 or self._stream is None:
            raise NullPointerDereference

        if self._value is None and not issubclass(self.type, Void):
            # Read current position of file read/write pointer
            position = self._stream.tell()
            # Reposition the file read/write pointer
            self._stream.seek(self)

            if issubclass(self.type, Char):
                # this makes the assumption that a char pointer is a null-terminated string
                value = self.type._read_0(self._stream, self._context)
            else:
                value = self.type._read(self._stream, self._context)

            self._stream.seek(position)
            self._value = value

        return self._value



from dissect.cstruct.bitbuffer import BitBuffer
from dissect.cstruct.cstruct import cstruct, ctypes, ctypes_type
from dissect.cstruct.exceptions import (
    Error,
    NullPointerDereference,
    ParserError,
    ResolveError,
)
from dissect.cstruct.expression import Expression
from dissect.cstruct.types import (
    LEB128,
    Array,
    BaseType,
    Char,
    CharArray,
    Enum,
    Field,
    Flag,
    Int,
    MetaType,
    Packed,
    Pointer,
    Structure,
    Union,
    Void,
    Wchar,
    WcharArray,
)
from dissect.cstruct.utils import (
    dumpstruct,
    hexdump,
    p8,
    p16,
    p32,
    p64,
    pack,
    swap,
    swap16,
    swap32,
    swap64,
    u8,
    u16,
    u32,
    u64,
    unpack,
)

__all__ = [
    "LEB128",
    "Array",
    "BaseType",
    "BitBuffer",
    "Char",
    "CharArray",
    "Enum",
    "Error",
    "Expression",
    "Field",
    "Flag",
    "Int",
    "MetaType",
    "NullPointerDereference",
    "Packed",
    "ParserError",
    "Pointer",
    "ResolveError",
    "Structure",
    "Union",
    "Void",
    "Wchar",
    "WcharArray",
    "cstruct",
    "ctypes",
    "ctypes_type",
    "dumpstruct",
    "hexdump",
    "p8",
    "p16",
    "p32",
    "p64",
    "pack",
    "swap",
    "swap16",
    "swap32",
    "swap64",
    "u8",
    "u16",
    "u32",
    "u64",
    "unpack",
]



from __future__ import annotations

import pprint
import string
import sys
from enum import Enum
from typing import TYPE_CHECKING

from dissect.cstruct.types.pointer import Pointer
from dissect.cstruct.types.structure import Structure

if TYPE_CHECKING:
    from collections.abc import Iterator
    from typing import Literal

COLOR_RED = "\033[1;31m"
COLOR_GREEN = "\033[1;32m"
COLOR_YELLOW = "\033[1;33m"
COLOR_BLUE = "\033[1;34m"
COLOR_PURPLE = "\033[1;35m"
COLOR_CYAN = "\033[1;36m"
COLOR_WHITE = "\033[1;37m"
COLOR_NORMAL = "\033[1;0m"

COLOR_BG_RED = "\033[1;41m\033[1;37m"
COLOR_BG_GREEN = "\033[1;42m\033[1;37m"
COLOR_BG_YELLOW = "\033[1;43m\033[1;37m"
COLOR_BG_BLUE = "\033[1;44m\033[1;37m"
COLOR_BG_PURPLE = "\033[1;45m\033[1;37m"
COLOR_BG_CYAN = "\033[1;46m\033[1;37m"
COLOR_BG_WHITE = "\033[1;47m\033[1;30m"

PRINTABLE = string.digits + string.ascii_letters + string.punctuation + " "

ENDIANNESS_MAP: dict[str, Literal["big", "little"]] = {
    "@": sys.byteorder,
    "=": sys.byteorder,
    "<": "little",
    ">": "big",
    "!": "big",
    "network": "big",
}

Palette = list[tuple[int, str]]


def _hexdump(data: bytes, palette: Palette | None = None, offset: int = 0, prefix: str = "") -> Iterator[str]:
    """Hexdump some data.

    Args:
        data: Bytes to hexdump.
        offset: Byte offset of the hexdump.
        prefix: Optional prefix.
        palette: Colorize the hexdump using this color pattern.
    """
    if palette:
        palette = palette[::-1]

    remaining = 0
    active = None

    for i in range(0, len(data), 16):
        values = ""
        chars = []

        for j in range(16):
            if not active and palette:
                remaining, active = palette.pop()
                while remaining == 0:
                    if len(palette) == 0:
                        # Last palette tuple is empty: print remaining whitespaces
                        active = ""
                        break
                    else:
                        remaining, active = palette.pop()
                values += active
            elif active and j == 0:
                values += active

            if i + j >= len(data):
                values += "  "
            else:
                char = data[i + j]
                char = chr(char)

                print_char = char if char in PRINTABLE else "."

                if active:
                    values += f"{ord(char):02x}"
                    chars.append(active + print_char + COLOR_NORMAL)
                else:
                    values += f"{ord(char):02x}"
                    chars.append(print_char)

                remaining -= 1
                if remaining == 0:
                    active = None

                    if palette is not None:
                        values += COLOR_NORMAL

                if j == 15 and palette is not None:
                    values += COLOR_NORMAL

            values += " "
            if j == 7:
                values += " "

        chars = "".join(chars)
        yield f"{prefix}{offset + i:08x}  {values:48s}  {chars}"


def hexdump(
    data: bytes, palette: Palette | None = None, offset: int = 0, prefix: str = "", output: str = "print"
) -> Iterator[str] | str | None:
    """Hexdump some data.

    Args:
        data: Bytes to hexdump.
        palette: Colorize the hexdump using this color pattern.
        offset: Byte offset of the hexdump.
        prefix: Optional prefix.
        output: Output format, can be 'print', 'generator' or 'string'.
    """
    generator = _hexdump(data, palette, offset, prefix)
    if output == "print":
        print("\n".join(generator))
        return None
    if output == "generator":
        return generator
    if output == "string":
        return "\n".join(list(generator))
    raise ValueError(f"Invalid output argument: {output!r} (should be 'print', 'generator' or 'string').")


def _dumpstruct(
    structure: Structure,
    data: bytes,
    offset: int,
    color: bool,
    output: str,
) -> str | None:
    palette = []
    colors = [
        (COLOR_RED, COLOR_BG_RED),
        (COLOR_GREEN, COLOR_BG_GREEN),
        (COLOR_YELLOW, COLOR_BG_YELLOW),
        (COLOR_BLUE, COLOR_BG_BLUE),
        (COLOR_PURPLE, COLOR_BG_PURPLE),
        (COLOR_CYAN, COLOR_BG_CYAN),
        (COLOR_WHITE, COLOR_BG_WHITE),
    ]
    ci = 0
    out = [f"struct {structure.__class__.__name__}:"]
    foreground, background = None, None
    for field in structure.__class__.__fields__:
        if getattr(field.type, "anonymous", False):
            continue

        value = getattr(structure, field._name)
        if isinstance(value, (str, Pointer, Enum)):
            value = repr(value)
        elif isinstance(value, int):
            value = hex(value)
        elif isinstance(value, list):
            value = pprint.pformat(value)
            if "\n" in value:
                value = value.replace("\n", f"\n{' ' * (len(field._name) + 4)}")

        if color:
            foreground, background = colors[ci % len(colors)]
            size = structure.__sizes__[field._name]
            palette.append((size, background))
            ci += 1
            out.append(f"- {foreground}{field._name}{COLOR_NORMAL}: {value}")
        else:
            out.append(f"- {field._name}: {value}")

    out = "\n".join(out)

    if output == "print":
        print()
        hexdump(data, palette, offset=offset)
        print()
        print(out)
    elif output == "string":
        return f"\n{hexdump(data, palette, offset=offset, output='string')}\n\n{out}"
    return None


def dumpstruct(
    obj: Structure | type[Structure],
    data: bytes | None = None,
    offset: int = 0,
    color: bool = True,
    output: str = "print",
) -> str | None:
    """Dump a structure or parsed structure instance.

    Prints a colorized hexdump and parsed structure output.

    Args:
        obj: Structure to dump.
        data: Bytes to parse the Structure on, if obj is not a parsed Structure already.
        offset: Byte offset of the hexdump.
        output: Output format, can be 'print' or 'string'.
    """
    if output not in ("print", "string"):
        raise ValueError(f"Invalid output argument: {output!r} (should be 'print' or 'string').")

    if isinstance(obj, Structure):
        return _dumpstruct(obj, obj.dumps(), offset, color, output)
    if issubclass(obj, Structure) and data is not None:
        return _dumpstruct(obj(data), data, offset, color, output)
    raise ValueError("Invalid arguments")


def pack(value: int, size: int | None = None, endian: str = "little") -> bytes:
    """Pack an integer value to a given bit size, endianness.

    Arguments:
        value: Value to pack.
        size: Integer size in bits.
        endian: Endianness to use (little, big, network, <, > or !)
    """
    size = ((size or value.bit_length()) + 7) // 8
    return value.to_bytes(size, ENDIANNESS_MAP.get(endian, endian), signed=value < 0)


def unpack(value: bytes, size: int | None = None, endian: str = "little", sign: bool = False) -> int:
    """Unpack an integer value from a given bit size, endianness and sign.

    Arguments:
        value: Value to unpack.
        size: Integer size in bits.
        endian: Endianness to use (little, big, network, <, > or !)
        sign: Signedness of the integer.
    """
    if size and len(value) != size // 8:
        raise ValueError(f"Invalid byte value, expected {size // 8} bytes, got {len(value)} bytes")
    return int.from_bytes(value, ENDIANNESS_MAP.get(endian, endian), signed=sign)


def p8(value: int, endian: str = "little") -> bytes:
    """Pack an 8 bit integer.

    Arguments:
        value: Value to pack.
        endian: Endianness to use (little, big, network, <, > or !)
    """
    return pack(value, 8, endian)


def p16(value: int, endian: str = "little") -> bytes:
    """Pack a 16 bit integer.

    Arguments:
        value: Value to pack.
        endian: Endianness to use (little, big, network, <, > or !)
    """
    return pack(value, 16, endian)


def p32(value: int, endian: str = "little") -> bytes:
    """Pack a 32 bit integer.

    Arguments:
        value: Value to pack.
        endian: Endianness to use (little, big, network, <, > or !)
    """
    return pack(value, 32, endian)


def p64(value: int, endian: str = "little") -> bytes:
    """Pack a 64 bit integer.

    Arguments:
        value: Value to pack.
        endian: Endianness to use (little, big, network, <, > or !)
    """
    return pack(value, 64, endian)


def u8(value: bytes, endian: str = "little", sign: bool = False) -> int:
    """Unpack an 8 bit integer.

    Arguments:
        value: Value to unpack.
        endian: Endianness to use (little, big, network, <, > or !)
        sign: Signedness of the integer.
    """
    return unpack(value, 8, endian, sign)


def u16(value: bytes, endian: str = "little", sign: bool = False) -> int:
    """Unpack a 16 bit integer.

    Arguments:
        value: Value to unpack.
        endian: Endianness to use (little, big, network, <, > or !)
        sign: Signedness of the integer.
    """
    return unpack(value, 16, endian, sign)


def u32(value: bytes, endian: str = "little", sign: bool = False) -> int:
    """Unpack a 32 bit integer.

    Arguments:
        value: Value to unpack.
        endian: Endianness to use (little, big, network, <, > or !)
        sign: Signedness of the integer.
    """
    return unpack(value, 32, endian, sign)


def u64(value: bytes, endian: str = "little", sign: bool = False) -> int:
    """Unpack a 64 bit integer.

    Arguments:
        value: Value to unpack.
        endian: Endianness to use (little, big, network, <, > or !)
        sign: Signedness of the integer.
    """
    return unpack(value, 64, endian, sign)


def swap(value: int, size: int) -> int:
    """Swap the endianness of an integer with a given bit size.

    Arguments:
        value: Integer to swap.
        size: Integer size in bits.
    """
    return unpack(pack(value, size, ">"), size, "<")


def swap16(value: int) -> int:
    """Swap the endianness of a 16 bit integer.

    Arguments:
        value: Integer to swap.
    """
    return swap(value, 16)


def swap32(value: int) -> int:
    """Swap the endianness of a 32 bit integer.

    Arguments:
        value: Integer to swap.
    """
    return swap(value, 32)


def swap64(value: int) -> int:
    """Swap the endianness of a 64 bit integer.

    Arguments:
        value: Integer to swap.
    """
    return swap(value, 64)



from __future__ import annotations

import sys
from typing import Any, BinaryIO, ClassVar

from dissect.cstruct.types.base import EOF, BaseArray, BaseType


class WcharArray(str, BaseArray):
    """Wide-character array type for reading and writing UTF-16 strings."""

    __slots__ = ()

    @classmethod
    def __default__(cls) -> WcharArray:
        return type.__call__(cls, "\x00" * (0 if cls.dynamic or cls.null_terminated else cls.num_entries))

    @classmethod
    def _read(cls, stream: BinaryIO, context: dict[str, Any] | None = None) -> WcharArray:
        return type.__call__(cls, super()._read(stream, context))

    @classmethod
    def _write(cls, stream: BinaryIO, data: str) -> int:
        if cls.null_terminated:
            data += "\x00"
        return stream.write(data.encode(Wchar.__encoding_map__[cls.cs.endian]))


class Wchar(str, BaseType):
    """Wide-character type for reading and writing UTF-16 characters."""

    ArrayType = WcharArray

    __slots__ = ()
    __encoding_map__: ClassVar[dict[str, str]] = {
        "@": f"utf-16-{sys.byteorder[0]}e",
        "=": f"utf-16-{sys.byteorder[0]}e",
        "<": "utf-16-le",
        ">": "utf-16-be",
        "!": "utf-16-be",
    }

    @classmethod
    def __default__(cls) -> Wchar:
        return type.__call__(cls, "\x00")

    @classmethod
    def _read(cls, stream: BinaryIO, context: dict[str, Any] | None = None) -> Wchar:
        return cls._read_array(stream, 1, context)

    @classmethod
    def _read_array(cls, stream: BinaryIO, count: int, context: dict[str, Any] | None = None) -> Wchar:
        if count == 0:
            return type.__call__(cls, "")

        if count != EOF:
            count *= 2

        data = stream.read(-1 if count == EOF else count)
        if count != EOF and len(data) != count:
            raise EOFError(f"Read {len(data)} bytes, but expected {count}")

        return type.__call__(cls, data.decode(cls.__encoding_map__[cls.cs.endian]))

    @classmethod
    def _read_0(cls, stream: BinaryIO, context: dict[str, Any] | None = None) -> Wchar:
        buf = []
        while True:
            point = stream.read(2)
            if (bytes_read := len(point)) != 2:
                raise EOFError(f"Read {bytes_read} bytes, but expected 2")

            if point == b"\x00\x00":
                break

            buf.append(point)

        return type.__call__(cls, b"".join(buf).decode(cls.__encoding_map__[cls.cs.endian]))

    @classmethod
    def _write(cls, stream: BinaryIO, data: str) -> int:
        return stream.write(data.encode(cls.__encoding_map__[cls.cs.endian]))



from __future__ import annotations

from functools import lru_cache
from struct import Struct
from typing import TYPE_CHECKING, Any, BinaryIO, Generic, TypeVar

from dissect.cstruct.types.base import EOF, BaseType

if TYPE_CHECKING:
    from typing_extensions import Self


@lru_cache(1024)
def _struct(endian: str, packchar: str) -> Struct:
    return Struct(f"{endian}{packchar}")


T = TypeVar("T", int, float)


class Packed(BaseType, Generic[T]):
    """Packed type for Python struct (un)packing."""

    packchar: str

    @classmethod
    def _read(cls, stream: BinaryIO, context: dict[str, Any] | None = None) -> Self:
        return cls._read_array(stream, 1, context)[0]

    @classmethod
    def _read_array(cls, stream: BinaryIO, count: int, context: dict[str, Any] | None = None) -> list[Self]:
        if count == EOF:
            data = stream.read()
            length = len(data)
            count = length // cls.size
        else:
            length = cls.size * count
            data = stream.read(length)

        fmt = _struct(cls.cs.endian, f"{count}{cls.packchar}")

        if len(data) != length:
            raise EOFError(f"Read {len(data)} bytes, but expected {length}")

        return [cls.__new__(cls, value) for value in fmt.unpack(data)]

    @classmethod
    def _read_0(cls, stream: BinaryIO, context: dict[str, Any] | None = None) -> Self:
        result = []

        fmt = _struct(cls.cs.endian, cls.packchar)
        while True:
            data = stream.read(cls.size)

            if len(data) != cls.size:
                raise EOFError(f"Read {len(data)} bytes, but expected {cls.size}")

            if (value := fmt.unpack(data)[0]) == 0:
                break

            result.append(cls.__new__(cls, value))

        return result

    @classmethod
    def _write(cls, stream: BinaryIO, data: Packed[T]) -> int:
        return stream.write(_struct(cls.cs.endian, cls.packchar).pack(data))

    @classmethod
    def _write_array(cls, stream: BinaryIO, data: list[Packed[T]]) -> int:
        return stream.write(_struct(cls.cs.endian, f"{len(data)}{cls.packchar}").pack(*data))



from __future__ import annotations

import ast
import re
from typing import TYPE_CHECKING, Any

from dissect.cstruct import compiler
from dissect.cstruct.exceptions import (
    ExpressionParserError,
    ExpressionTokenizerError,
    ParserError,
)
from dissect.cstruct.expression import Expression
from dissect.cstruct.types import BaseArray, BaseType, Field, Structure

if TYPE_CHECKING:
    from dissect.cstruct import cstruct


class Parser:
    """Base class for definition parsers.

    Args:
        cs: An instance of cstruct.
    """

    def __init__(self, cs: cstruct):
        self.cstruct = cs

    def parse(self, data: str) -> None:
        """This function should parse definitions to cstruct types.

        Args:
            data: Data to parse definitions from, usually a string.
        """
        raise NotImplementedError


class TokenParser(Parser):
    """
    Args:
        cs: An instance of cstruct.
        compiled: Whether structs should be compiled or not.
    """

    def __init__(self, cs: cstruct, compiled: bool = True, align: bool = False):
        super().__init__(cs)

        self.compiled = compiled
        self.align = align
        self.TOK = self._tokencollection()
        self._conditionals = []
        self._conditionals_depth = 0

    @staticmethod
    def _tokencollection() -> TokenCollection:
        TOK = TokenCollection()
        TOK.add(r"#\[(?P<values>[^\]]+)\](?=\s*)", "CONFIG_FLAG")
        TOK.add(r"#define\s+(?P<name>[^\s]+)(?P<value>[^\r\n]*)", "DEFINE")
        TOK.add(r"#undef\s+(?P<name>[^\s]+)\s*", "UNDEF")
        TOK.add(r"#ifdef\s+(?P<name>[^\s]+)\s*", "IFDEF")
        TOK.add(r"#ifndef\s+(?P<name>[^\s]+)\s*", "IFNDEF")
        TOK.add(r"#else\s*", "ELSE")
        TOK.add(r"#endif\s*", "ENDIF")
        TOK.add(r"typedef(?=\s)", "TYPEDEF")
        TOK.add(r"(?:struct|union)(?=\s|{)", "STRUCT")
        TOK.add(
            r"(?P<enumtype>enum|flag)\s+(?P<name>[^\s:{]+)?\s*(:\s"
            r"*(?P<type>[^{]+?)\s*)?\{(?P<values>[^}]+)\}\s*(?=;)",
            "ENUM",
        )
        TOK.add(r"(?<=})\s*(?P<defs>(?:[a-zA-Z0-9_]+\s*,\s*)+[a-zA-Z0-9_]+)\s*(?=;)", "DEFS")
        TOK.add(r"(?P<name>\**?\s*[a-zA-Z0-9_]+)(?:\s*:\s*(?P<bits>\d+))?(?:\[(?P<count>[^;]*)\])?\s*(?=;)", "NAME")
        TOK.add(r"#include\s+(?P<name>[^\s]+)\s*", "INCLUDE")
        TOK.add(r"[a-zA-Z_][a-zA-Z0-9_]*", "IDENTIFIER")
        TOK.add(r"[{}]", "BLOCK")
        TOK.add(r"\$(?P<name>[^\s]+) = (?P<value>{[^}]+})\w*[\r\n]+", "LOOKUP")
        TOK.add(r";", "EOL")
        TOK.add(r"\s+", None)
        TOK.add(r".", None)

        return TOK

    def _identifier(self, tokens: TokenConsumer) -> str:
        idents = []
        while tokens.next == self.TOK.IDENTIFIER:
            idents.append(tokens.consume())
        return " ".join([i.value for i in idents])

    def _conditional(self, tokens: TokenConsumer) -> None:
        token = tokens.consume()
        pattern = self.TOK.patterns[token.token]
        match = pattern.match(token.value).groupdict()

        value = match["name"]

        if token.token == self.TOK.IFDEF:
            self._conditionals.append(value in self.cstruct.consts)
        elif token.token == self.TOK.IFNDEF:
            self._conditionals.append(value not in self.cstruct.consts)

    def _check_conditional(self, tokens: TokenConsumer) -> bool:
        """Check and handle conditionals. Return a boolean indicating if we need to continue to the next token."""
        if self._conditionals and self._conditionals_depth == len(self._conditionals):
            # If we have a conditional and the depth matches, handle it accordingly
            if tokens.next == self.TOK.ELSE:
                # Flip the last conditional
                tokens.consume()
                self._conditionals[-1] = not self._conditionals[-1]
                return True

            if tokens.next == self.TOK.ENDIF:
                # Pop the last conditional
                tokens.consume()
                self._conditionals.pop()
                self._conditionals_depth -= 1
                return True

        if tokens.next in (self.TOK.IFDEF, self.TOK.IFNDEF):
            # If we encounter a new conditional, increase the depth
            self._conditionals_depth += 1

        if tokens.next == self.TOK.ENDIF:
            # Similarly, decrease the depth if needed
            self._conditionals_depth -= 1

        if self._conditionals and not self._conditionals[-1]:
            # If the last conditional evaluated to False, skip the next token
            tokens.consume()
            return True

        if tokens.next in (self.TOK.IFDEF, self.TOK.IFNDEF):
            # If the next token is a conditional, process it
            self._conditional(tokens)
            return True

        return False

    def _constant(self, tokens: TokenConsumer) -> None:
        const = tokens.consume()
        pattern = self.TOK.patterns[self.TOK.DEFINE]
        match = pattern.match(const.value).groupdict()

        value = match["value"].strip()
        try:
            value = ast.literal_eval(value)
        except (ValueError, SyntaxError):
            pass

        if isinstance(value, str):
            try:
                value = Expression(value).evaluate(self.cstruct)
            except (ExpressionParserError, ExpressionTokenizerError):
                pass

        self.cstruct.consts[match["name"]] = value

    def _undef(self, tokens: TokenConsumer) -> None:
        const = tokens.consume()
        pattern = self.TOK.patterns[self.TOK.UNDEF]
        match = pattern.match(const.value).groupdict()

        if match["name"] in self.cstruct.consts:
            del self.cstruct.consts[match["name"]]
        else:
            raise ParserError(f"line {self._lineno(const)}: constant {match['name']!r} not defined")

    def _enum(self, tokens: TokenConsumer) -> None:
        # We cheat with enums because the entire enum is in the token
        etok = tokens.consume()

        pattern = self.TOK.patterns[self.TOK.ENUM]
        # Dirty trick because the regex expects a ; but we don't want it to be part of the value
        d = pattern.match(etok.value + ";").groupdict()
        enumtype = d["enumtype"]

        nextval = 0
        if enumtype == "flag":
            nextval = 1

        values = {}
        for line in d["values"].splitlines():
            for v in line.split(","):
                key, _, val = v.partition("=")
                key = key.strip()
                val = val.strip()
                if not key:
                    continue

                val = nextval if not val else Expression(val).evaluate(self.cstruct, values)

                if enumtype == "flag":
                    high_bit = val.bit_length() - 1
                    nextval = 2 ** (high_bit + 1)
                else:
                    nextval = val + 1

                values[key] = val

        if not d["type"]:
            d["type"] = "uint32"

        factory = self.cstruct._make_flag if enumtype == "flag" else self.cstruct._make_enum

        enum = factory(d["name"] or "", self.cstruct.resolve(d["type"]), values)
        if not enum.__name__:
            self.cstruct.consts.update(enum.__members__)
        else:
            self.cstruct.add_type(enum.__name__, enum)

        tokens.eol()

    def _typedef(self, tokens: TokenConsumer) -> None:
        tokens.consume()
        type_ = None

        names = []

        if tokens.next == self.TOK.IDENTIFIER:
            type_ = self.cstruct.resolve(self._identifier(tokens))
        elif tokens.next == self.TOK.STRUCT:
            type_ = self._struct(tokens)
            if not type_.__anonymous__:
                names.append(type_.__name__)

        names.extend(self._names(tokens))
        for name in names:
            if issubclass(type_, Structure) and type_.__anonymous__:
                type_.__anonymous__ = False
                type_.__name__ = name
                type_.__qualname__ = name

            type_, name, bits = self._parse_field_type(type_, name)
            if bits is not None:
                raise ParserError(f"line {self._lineno(tokens.previous)}: typedefs cannot have bitfields")
            self.cstruct.add_type(name, type_)

    def _struct(self, tokens: TokenConsumer, register: bool = False) -> type[Structure]:
        stype = tokens.consume()

        factory = self.cstruct._make_union if stype.value.startswith("union") else self.cstruct._make_struct

        st = None
        names = []
        registered = False

        if tokens.next == self.TOK.IDENTIFIER:
            ident = tokens.consume()
            if register:
                # Pre-register an empty struct for self-referencing
                # We update this instance later with the fields
                st = factory(ident.value, [], align=self.align)
                if self.compiled and "nocompile" not in tokens.flags:
                    st = compiler.compile(st)
                self.cstruct.add_type(ident.value, st)
                registered = True
            else:
                names.append(ident.value)

        if tokens.next == self.TOK.NAME:
            # As part of a struct field
            # struct type_name field_name;
            if not names:
                raise ParserError(f"line {self._lineno(tokens.next)}: unexpected anonymous struct")
            return self.cstruct.resolve(names[0])

        if tokens.next != self.TOK.BLOCK:
            raise ParserError(f"line {self._lineno(tokens.next)}: expected start of block '{tokens.next}'")

        fields = []
        tokens.consume()
        while len(tokens):
            if tokens.next == self.TOK.BLOCK and tokens.next.value == "}":
                tokens.consume()
                break

            if self._check_conditional(tokens):
                continue

            field = self._parse_field(tokens)
            fields.append(field)

        if register:
            names.extend(self._names(tokens))

        # If the next token is EOL, consume it
        # Otherwise we're part of a typedef or field definition
        if tokens.next == self.TOK.EOL:
            tokens.eol()

        name = names[0] if names else None

        if st is None:
            is_anonymous = False
            if not name:
                is_anonymous = True
                name = self.cstruct._next_anonymous()

            st = factory(name, fields, align=self.align, anonymous=is_anonymous)
            if self.compiled and "nocompile" not in tokens.flags:
                st = compiler.compile(st)
        else:
            st.__fields__.extend(fields)
            st.commit()

        # This is pretty dirty
        if register:
            if not names and not registered:
                raise ParserError(f"line {self._lineno(stype)}: struct has no name")

            for name in names:
                self.cstruct.add_type(name, st)

        tokens.reset_flags()
        return st

    def _lookup(self, tokens: TokenConsumer) -> None:
        # Just like enums, we cheat and have the entire lookup in the token
        ltok = tokens.consume()

        pattern = self.TOK.patterns[self.TOK.LOOKUP]
        # Dirty trick because the regex expects a ; but we don't want it to be part of the value
        m = pattern.match(ltok.value + ";")
        d = ast.literal_eval(m.group(2))
        self.cstruct.lookups[m.group(1)] = {self.cstruct.consts[k]: v for k, v in d.items()}

    def _parse_field(self, tokens: TokenConsumer) -> Field:
        type_ = None
        if tokens.next == self.TOK.IDENTIFIER:
            type_ = self.cstruct.resolve(self._identifier(tokens))
        elif tokens.next == self.TOK.STRUCT:
            type_ = self._struct(tokens)

            if tokens.next != self.TOK.NAME:
                return Field(None, type_, None)

        if tokens.next != self.TOK.NAME:
            raise ParserError(f"line {self._lineno(tokens.next)}: expected name, got {tokens.next!r}")
        nametok = tokens.consume()

        type_, name, bits = self._parse_field_type(type_, nametok.value)

        tokens.eol()
        return Field(name.strip(), type_, bits)

    def _parse_field_type(self, type_: type[BaseType], name: str) -> tuple[type[BaseType], str, int | None]:
        pattern = self.TOK.patterns[self.TOK.NAME]
        # Dirty trick because the regex expects a ; but we don't want it to be part of the value
        d = pattern.match(name + ";").groupdict()

        name = d["name"]
        count_expression = d["count"]

        while name.startswith("*"):
            name = name[1:]
            type_ = self.cstruct._make_pointer(type_)

        if count_expression is not None:
            # Poor mans multi-dimensional array by abusing the eager regex match of count
            counts = count_expression.split("][") if "][" in count_expression else [count_expression]

            for count in reversed(counts):
                if count == "":
                    count = None
                else:
                    count = Expression(count)
                    try:
                        count = count.evaluate(self.cstruct)
                    except Exception:
                        pass

                if issubclass(type_, BaseArray) and count is None:
                    raise ParserError("Depth required for multi-dimensional array")

                type_ = self.cstruct._make_array(type_, count)

        return type_, name.strip(), int(d["bits"]) if d["bits"] else None

    def _names(self, tokens: TokenConsumer) -> list[str]:
        names = []
        while True:
            if tokens.next == self.TOK.EOL:
                tokens.eol()
                break

            if tokens.next not in (self.TOK.NAME, self.TOK.DEFS, self.TOK.IDENTIFIER):
                break

            ntoken = tokens.consume()
            if ntoken in (self.TOK.NAME, self.TOK.IDENTIFIER):
                names.append(ntoken.value.strip())
            elif ntoken == self.TOK.DEFS:
                names.extend([name.strip() for name in ntoken.value.strip().split(",")])

        return names

    def _include(self, tokens: TokenConsumer) -> None:
        include = tokens.consume()
        pattern = self.TOK.patterns[self.TOK.INCLUDE]
        match = pattern.match(include.value).groupdict()

        self.cstruct.includes.append(match["name"].strip().strip("'\""))

    @staticmethod
    def _remove_comments(string: str) -> str:
        # https://stackoverflow.com/a/18381470
        pattern = r"(\".*?\"|\'.*?\')|(/\*.*?\*/|//[^\r\n]*$)"
        # first group captures quoted strings (double or single)
        # second group captures comments (//single-line or /* multi-line */)
        regex = re.compile(pattern, re.MULTILINE | re.DOTALL)

        def _replacer(match: re.Match) -> str:
            # if the 2nd group (capturing comments) is not None,
            # it means we have captured a non-quoted (real) comment string.
            if comment := match.group(2):
                return "\n" * comment.count("\n")  # so we will return empty to remove the comment
            # otherwise, we will return the 1st group
            return match.group(1)  # captured quoted-string

        return regex.sub(_replacer, string)

    @staticmethod
    def _lineno(tok: Token) -> int:
        """Quick and dirty line number calculator"""

        match = tok.match
        return match.string.count("\n", 0, match.start()) + 1

    def _config_flag(self, tokens: TokenConsumer) -> None:
        flag_token = tokens.consume()
        pattern = self.TOK.patterns[self.TOK.CONFIG_FLAG]
        tok_dict = pattern.match(flag_token.value).groupdict()
        tokens.flags.extend(tok_dict["values"].split(","))

    def parse(self, data: str) -> None:
        scanner = re.Scanner(self.TOK.tokens)
        data = self._remove_comments(data)
        tokens, remaining = scanner.scan(data)

        if len(remaining):
            lineno = data.count("\n", 0, len(data) - len(remaining))
            raise ParserError(f"line {lineno}: invalid syntax in definition")

        tokens = TokenConsumer(tokens)
        while True:
            token = tokens.next
            if token is None:
                break

            if self._check_conditional(tokens):
                continue

            if token == self.TOK.CONFIG_FLAG:
                self._config_flag(tokens)
            elif token == self.TOK.DEFINE:
                self._constant(tokens)
            elif token == self.TOK.UNDEF:
                self._undef(tokens)
            elif token == self.TOK.TYPEDEF:
                self._typedef(tokens)
            elif token == self.TOK.STRUCT:
                self._struct(tokens, register=True)
            elif token == self.TOK.ENUM:
                self._enum(tokens)
            elif token == self.TOK.LOOKUP:
                self._lookup(tokens)
            elif token == self.TOK.INCLUDE:
                self._include(tokens)
            else:
                raise ParserError(f"line {self._lineno(token)}: unexpected token {token!r}")

        if self._conditionals:
            raise ParserError(f"line {self._lineno(tokens.previous)}: unclosed conditional statement")


class CStyleParser(Parser):
    """Definition parser for C-like structure syntax.

    Args:
        cs: An instance of cstruct
        compiled: Whether structs should be compiled or not.
    """

    def __init__(self, cs: cstruct, compiled: bool = True):
        self.compiled = compiled
        super().__init__(cs)

    def _constants(self, data: str) -> None:
        r = re.finditer(r"#define\s+(?P<name>[^\s]+)\s+(?P<value>[^\r\n]+)\s*\n", data)
        for t in r:
            d = t.groupdict()
            v = d["value"].rsplit("//")[0]

            try:
                v = ast.literal_eval(v)
            except (ValueError, SyntaxError):
                pass

            self.cstruct.consts[d["name"]] = v

    def _enums(self, data: str) -> None:
        r = re.finditer(
            r"(?P<enumtype>enum|flag)\s+(?P<name>[^\s:{]+)\s*(:\s*(?P<type>[^\s]+)\s*)?\{(?P<values>[^}]+)\}\s*;",
            data,
        )
        for t in r:
            d = t.groupdict()
            enumtype = d["enumtype"]

            nextval = 0
            if enumtype == "flag":
                nextval = 1

            values = {}
            for line in d["values"].split("\n"):
                line, _, _ = line.partition("//")
                for v in line.split(","):
                    key, _, val = v.partition("=")
                    key = key.strip()
                    val = val.strip()
                    if not key:
                        continue

                    val = nextval if not val else Expression(val).evaluate(self.cstruct)

                    if enumtype == "flag":
                        high_bit = val.bit_length() - 1
                        nextval = 2 ** (high_bit + 1)
                    else:
                        nextval = val + 1

                    values[key] = val

            if not d["type"]:
                d["type"] = "uint32"

            factory = self.cstruct._make_enum
            if enumtype == "flag":
                factory = self.cstruct._make_flag

            enum = factory(d["name"], self.cstruct.resolve(d["type"]), values)
            self.cstruct.add_type(enum.__name__, enum)

    def _structs(self, data: str) -> None:
        r = re.finditer(
            r"(#(?P<flags>(?:compile))\s+)?"
            r"((?P<typedef>typedef)\s+)?"
            r"(?P<type>[^\s]+)\s+"
            r"(?P<name>[^\s]+)?"
            r"(?P<fields>"
            r"\s*{[^}]+\}(?P<defs>\s+[^;\n]+)?"
            r")?\s*;",
            data,
        )
        for t in r:
            d = t.groupdict()

            if d["name"]:
                name = d["name"]
            elif d["defs"]:
                name = d["defs"].strip().split(",")[0].strip()
            else:
                raise ParserError("No name for struct")

            if d["type"] == "struct":
                data = self._parse_fields(d["fields"][1:-1].strip())
                st = self.cstruct._make_struct(name, data)
                if d["flags"] == "compile" or self.compiled:
                    st = compiler.compile(st)
            elif d["typedef"] == "typedef":
                st = d["type"]
            else:
                continue

            if d["name"]:
                self.cstruct.add_type(d["name"], st)

            if d["defs"]:
                for td in d["defs"].strip().split(","):
                    td = td.strip()
                    self.cstruct.add_type(td, st)

    def _parse_fields(self, data: str) -> None:
        fields = re.finditer(
            r"(?P<type>[^\s]+)\s+(?P<name>[^\s\[:]+)(:(?P<bits>\d+))?(\[(?P<count>[^;\n]*)\])?;",
            data,
        )

        result = []
        for f in fields:
            d = f.groupdict()
            if d["type"].startswith("//"):
                continue

            type_ = self.cstruct.resolve(d["type"])

            d["name"] = d["name"].replace("(", "").replace(")", "")

            # Maybe reimplement lazy type references later
            # _type = TypeReference(self, d['type'])
            if d["count"] is not None:
                if d["count"] == "":
                    count = None
                else:
                    count = Expression(d["count"])
                    try:
                        count = count.evaluate(self.cstruct)
                    except Exception:
                        pass

                type_ = self.cstruct._make_array(type_, count)

            if d["name"].startswith("*"):
                d["name"] = d["name"][1:]
                type_ = self.cstruct._make_pointer(type_)

            field = Field(d["name"], type_, int(d["bits"]) if d["bits"] else None)
            result.append(field)

        return result

    def _lookups(self, data: str, consts: dict[str, int]) -> None:
        r = re.finditer(r"\$(?P<name>[^\s]+) = ({[^}]+})\w*\n", data)

        for t in r:
            d = ast.literal_eval(t.group(2))
            self.cstruct.lookups[t.group(1)] = {self.cstruct.consts[k]: v for k, v in d.items()}

    def parse(self, data: str) -> None:
        self._constants(data)
        self._enums(data)
        self._structs(data)
        self._lookups(data, self.cstruct.consts)


class Token:
    __slots__ = ("match", "token", "value")

    def __init__(self, token: str, value: str, match: re.Match):
        self.token = token
        self.value = value
        self.match = match

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Token):
            other = other.token

        return self.token == other

    def __ne__(self, other: object) -> bool:
        return not self == other

    def __repr__(self) -> str:
        return f"<Token.{self.token} value={self.value!r}>"


class TokenCollection:
    def __init__(self):
        self.tokens: list[Token] = []
        self.lookup: dict[str, str] = {}
        self.patterns: dict[str, re.Pattern] = {}

    def __getattr__(self, attr: str) -> str | Any:
        try:
            return self.lookup[attr]
        except AttributeError:
            pass

        return object.__getattribute__(self, attr)

    def add(self, regex: str, name: str | None) -> None:
        if name is None:
            self.tokens.append((regex, None))
        else:
            self.lookup[name] = name
            self.patterns[name] = re.compile(regex)
            self.tokens.append((regex, lambda s, t: Token(name, t, s.match)))


class TokenConsumer:
    def __init__(self, tokens: list[Token]):
        self.tokens = tokens
        self.flags = []
        self.previous = None

    def __contains__(self, token: Token) -> bool:
        return token in self.tokens

    def __len__(self) -> int:
        return len(self.tokens)

    def __repr__(self) -> str:
        return f"<TokenConsumer next={self.next!r}>"

    @property
    def next(self) -> Token:
        try:
            return self.tokens[0]
        except IndexError:
            return None

    def consume(self) -> Token:
        self.previous = self.tokens.pop(0)
        return self.previous

    def reset_flags(self) -> None:
        self.flags = []

    def eol(self) -> None:
        token = self.consume()
        if token.token != "EOL":
            raise ParserError(f"line {self._lineno(token)}: expected EOL")



from __future__ import annotations

from enum import IntFlag

from dissect.cstruct.types.base import BaseType
from dissect.cstruct.types.enum import PY_311, EnumMetaType


class Flag(BaseType, IntFlag, metaclass=EnumMetaType):
    """Flag type supercharged with cstruct functionality.

    Flags are (mostly) compatible with the Python 3 standard library ``IntFlag`` with some notable differences:
        - Flag members are only considered equal if the flag class is the same

    Flags can be made using any integer type.

    Example:
        When using the default C-style parser, the following syntax is supported::

            flag <name> [: <type>] {
                <values>
            };

        For example, a flag that has A=1, B=4 and C=8 could be written like so::

            flag Test : uint16 {
                A, B=4, C
            };
    """

    def __repr__(self) -> str:
        result = super().__repr__()
        if not self.__class__.__name__:
            # Deal with anonymous flags by stripping off the first bit
            # I.e. <.RED: 1> -> <RED: 1>
            result = f"<{result[2:]}"
        return result

    if PY_311:

        def __str__(self) -> str:
            # We differentiate with standard Python flags in that we use a more descriptive str representation
            # Standard Python flags just use the integer value as str, we use FlagName.ValueName
            # In case of anonymous flags, we just use the ValueName
            base = f"{self.__class__.__name__}." if self.__class__.__name__ else ""
            return f"{base}{self.name}"

    else:

        def __str__(self) -> str:
            result = IntFlag.__str__(self)
            if not self.__class__.__name__:
                # Deal with anonymous flags
                # I.e. .RED -> RED
                result = result[1:]
            return result

    def __eq__(self, other: int | Flag) -> bool:
        if isinstance(other, Flag) and other.__class__ is not self.__class__:
            return False

        # Python <= 3.10 compatibility
        if isinstance(other, Flag):
            other = other.value

        return self.value == other

    def __ne__(self, value: int | Flag) -> bool:
        return not self.__eq__(value)

    def __hash__(self) -> int:
        return hash((self.__class__, self.name, self.value))



from dissect.cstruct.types.base import Array, BaseArray, BaseType, MetaType
from dissect.cstruct.types.char import Char, CharArray
from dissect.cstruct.types.enum import Enum
from dissect.cstruct.types.flag import Flag
from dissect.cstruct.types.int import Int
from dissect.cstruct.types.leb128 import LEB128
from dissect.cstruct.types.packed import Packed
from dissect.cstruct.types.pointer import Pointer
from dissect.cstruct.types.structure import Field, Structure, Union
from dissect.cstruct.types.void import Void, VoidArray
from dissect.cstruct.types.wchar import Wchar, WcharArray

__all__ = [
    "LEB128",
    "Array",
    "BaseArray",
    "BaseType",
    "Char",
    "CharArray",
    "Enum",
    "Field",
    "Flag",
    "Int",
    "MetaType",
    "Packed",
    "Pointer",
    "Structure",
    "Union",
    "Void",
    "VoidArray",
    "Wchar",
    "WcharArray",
]
